import traceback


class TestListener(object):
    def on_test_suite_start(self, test_suite):
        pass

    def on_test_suite_finish(self, test_suite):
        pass

    def on_test_class_start(self, test_class):
        pass

    def on_test_class_finish(self, test_class):
        pass

    def on_test_group_start(self, test_group):
        pass

    def on_test_group_finish(self, test_group):
        pass

    def on_test_case_start(self, test_case):
        pass

    def on_test_case_finish(self, test_case):
        pass


class TestListenerGroup(object):
    def __init__(self):
        self.__test_listeners = []
        self.__teamcity_test_listener = None

    def append(self, test_listener):
        if test_listener.__class__.__name__ == "TeamcityTestListener" and test_listener.__class__.__module__ == "__main__":
            self.__teamcity_test_listener = test_listener
        else:
            self.__test_listeners.append(test_listener)

    def on_test_suite_start(self, test_suite):
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_suite_start(test_suite)
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_suite_start(test_suite)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))

    def on_test_suite_finish(self, test_suite):
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_suite_finish(test_suite)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_suite_finish(test_suite)

    def on_test_class_start(self, test_class):
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_class_start(test_class)
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_class_start(test_class)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))

    def on_test_class_finish(self, test_class):
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_class_finish(test_class)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_class_finish(test_class)

    def on_test_group_start(self, test_group):
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_group_start(test_group)
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_group_start(test_group)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))

    def on_test_group_finish(self, test_group):
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_group_finish(test_group)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_group_finish(test_group)

    def on_test_case_start(self, test_case):
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_case_start(test_case)
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_case_start(test_case)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))

    def on_test_case_finish(self, test_case):
        for test_listener in self.__test_listeners:
            try:
                test_listener.on_test_case_finish(test_case)
            except Exception:
                from .plogger import pconsole
                pconsole.write_line("The test listener %s.%s raised exception:\n%s"
                                    % (test_listener.__class__.__module__, test_listener.__class__.__name__, traceback.format_exc()))
        if self.__teamcity_test_listener:
            self.__teamcity_test_listener.on_test_case_finish(test_case)


test_listeners = TestListenerGroup()



class TestFilter:
    def filter(self, test_ref):
        return True

    def __str__(self):
        return "Filter Class: %s.%s" % (self.__module__, self.__class__.__name__)


class TestIncludeTagsFilter(TestFilter):
    def __init__(self, tags):
        self._tags = tags

    def filter(self, test_ref):
        return hasattr(test_ref, "__tags__") and len([val for val in self._tags if val in test_ref.__tags__]) != 0

    def __str__(self):
        return "Include Tags: %s" % ",".join(self._tags)


class TestExcludeTagsFilter(TestFilter):
    def __init__(self, tags):
        self._tags = tags

    def filter(self, test_ref):
        return hasattr(test_ref, "__tags__") and len([val for val in self._tags if val in test_ref.__tags__]) == 0

    def __str__(self):
        return "Exclude Tags: %s" % ",".join(self._tags)


class TestIncludeGroupsFilter(TestFilter):
    def __init__(self, groups):
        self._groups = groups

    def filter(self, test_ref):
        return hasattr(test_ref, "__group__") and test_ref.__group__ in self._groups

    def __str__(self):
        return "Include Groups: %s" % ",".join(self._groups)


class TestFilterGroup:
    def __init__(self):
        self.__filters = []

    def filter(self, test_ref):
        if not self.__filters:
            return True
        for ft in self.__filters:
            if not ft.filter(test_ref):
                return False
        return True

    def append_filter(self, test_filter):
        self.__filters.append(test_filter)

    def __len__(self):
        return len(self.__filters)

    def __getitem__(self, item):
        return self.__filters[item]



import ctypes
import errno
import os
import sys
import time
import traceback
import types

if sys.version_info[0] == 2:
    StringTypes = (str, unicode)
else:
    StringTypes = (str,)

if sys.version_info[0] == 2:
    from urlparse import urljoin
    from urllib import unquote, pathname2url
else:
    from urllib.parse import urljoin, unquote
    from urllib.request import pathname2url

def make_dirs(dir_path):
    try:
        os.makedirs(dir_path)
    except OSError as e:
        if e.errno != errno.EEXIST:  # be happy if someone already created the path
            raise


def remove_tree(dir_path, remove_root=True):
    if os.path.exists(dir_path):
        for root, dirs, files in os.walk(dir_path, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        if remove_root is True:
            os.rmdir(dir_path)


def is_coroutine_function(func):
    if sys.version_info[:2] >= (3, 4):
        import asyncio
        return asyncio.iscoroutinefunction(func)
    return False


def mock_func(func):
    fn = types.FunctionType(func.__code__, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
    # in case fun was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(func.__dict__)
    return fn


def call_function(func, *args, **kwargs):
    if is_coroutine_function(func):
        import asyncio
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(func.__call__(*args, **kwargs))
    return func.__call__(*args, **kwargs)


def kill_thread(thread):
    """
        Kill a python thread from another thread.

    :param thread: a threading.Thread instance
    """
    exc = ctypes.py_object(SystemExit)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(thread.ident), exc)
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")

    start_time = time.time()
    while (time.time() - start_time) <= 30:
        if not thread.isAlive():
            return
        time.sleep(1)

    raise SystemError("Timed out waiting for thread <%s> to be killed." % thread)


def format_thread_stack(thread):
    stack_code = ["Stack Trace:"]
    stack = sys._current_frames()[thread.ident]
    for file_name, line_no, name, line in traceback.extract_stack(stack):
        stack_code.append("  File: \"%s\", line %d, in %s" % (file_name, line_no, name))
        if line:
            stack_code.append("    %s" % (line.strip()))
    return "\n".join(stack_code)


def escape_html(obj):
    if isinstance(obj, dict):
        return {key: escape_html(value) for key, value in obj.items()}
    if isinstance(obj, list):
        return [escape_html(item) for item in obj]
    if isinstance(obj, StringTypes):
        return obj.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;") \
            .replace(" ", "&nbsp;").replace('"', "&quot;").replace("\n", "<br/>")
    return obj


def escape_filename(name):
    _name = name
    escape_list = [
        ("%", "%25"),
        ("\\", "%5C"),
        ("/", "%2F"),
        (":", "%3A"),
        ("*", "%01"),
        ("?", "%3F"),
        ("\"", "%22"),
        ("<", "%3C"),
        (">", "%3E"),
        ("|", "%7C")
    ]
    for char, to in escape_list:
        _name = _name.replace(char, to)
    return _name



import json
import os
import platform
import shutil
import traceback
from codecs import open
from datetime import datetime
from xml.dom import minidom

from . import config, __version__
from .plogger import pconsole
from .testsuite import default_test_suite
from .enumeration import TestCaseStatus, TestCaseCountItem
from .utils import make_dirs, remove_tree, escape_html

current_dir = os.path.dirname(os.path.abspath(__file__))


def generate_xunit_xml(xml_file_path):
    pconsole.write_line("Generating xunit report...")
    doc = minidom.Document()
    test_suite_ele = doc.createElement("testsuite")
    doc.appendChild(test_suite_ele)
    status_count = default_test_suite.status_count
    test_suite_ele.setAttribute("name", default_test_suite.name)
    test_suite_ele.setAttribute("tests", str(status_count[TestCaseCountItem.TOTAL]))
    test_suite_ele.setAttribute("failures", str(status_count[TestCaseCountItem.FAILED]))
    test_suite_ele.setAttribute("skips", str(status_count[TestCaseCountItem.SKIPPED]))
    test_suite_ele.setAttribute("errors", "0")
    test_suite_ele.setAttribute("time", "%.3f" % default_test_suite.elapsed_time)
    test_suite_ele.setAttribute("timestamp", str(default_test_suite.start_time))

    for test_case in default_test_suite.test_cases:
        test_case_ele = doc.createElement("testcase")
        test_suite_ele.appendChild(test_case_ele)
        test_case_ele.setAttribute("name", test_case.name)
        test_case_ele.setAttribute("classname", test_case.test_class.full_name)
        test_case_ele.setAttribute("time", "%.3f" % test_case.elapsed_time)
        if test_case.status == TestCaseStatus.SKIPPED:
            skipped_ele = doc.createElement("skipped")
            test_case_ele.appendChild(skipped_ele)
            skipped_ele.setAttribute("message", test_case.skip_message)
        elif test_case.status == TestCaseStatus.FAILED:
            failure_ele = doc.createElement("failure")
            test_case_ele.appendChild(failure_ele)
            failure_ele.setAttribute("message", test_case.failure_message)
            failure_ele.setAttribute("type", test_case.failure_type)
            failure_ele.appendChild(doc.createTextNode(test_case.stack_trace))

    if os.path.exists(xml_file_path):
        pconsole.write_line("Cleaning old xunit report...")
        os.remove(xml_file_path)
    else:
        make_dirs(os.path.dirname(xml_file_path))

    f = open(xml_file_path, mode="w", encoding="utf-8")
    try:
        doc.writexml(f, "\t", "\t", "\n", "utf-8")
        pconsole.write_line("xunit report is generated at %s" % xml_file_path)
    except Exception as e:
        pconsole.write_line("Failed to generate xunit report.\n%s" % traceback.format_exc())
    finally:
        f.close()


def generate_html_report(report_dir):
    pconsole.write_line("Generating html report...")

    if os.path.exists(report_dir):
        pconsole.write_line("Cleaning old html report...")
        remove_tree(report_dir, remove_root=False)
    else:
        make_dirs(report_dir)

    html_template_dir = os.path.join(current_dir, "htmltemplate")

    # copy js and css file to report dir
    for fn in os.listdir(html_template_dir):
        file_full_path = os.path.join(html_template_dir, fn)
        _, file_ext = os.path.splitext(fn)
        if os.path.isfile(file_full_path) and file_ext in [".js", ".css"]:
            shutil.copy(file_full_path, report_dir)

    # copy screenshot from temp dir
    temp_dir = config.get_option("temp")
    for fn in os.listdir(temp_dir):
        file_full_path = os.path.join(temp_dir, fn)
        _, file_ext = os.path.splitext(fn)
        if os.path.isfile(file_full_path) and file_ext == ".png":
            shutil.copy(file_full_path, report_dir)

    with open(os.path.join(html_template_dir, "index.html"), encoding="utf-8") as f:
        index_page_template = f.read()

    current_time = datetime.now()
    system_info = "%s / Python %s / %s" % (platform.node(), platform.python_version(), platform.platform())
    test_suite_json = json.dumps(_get_test_suite_dict(default_test_suite))
    index_page_content = index_page_template.format(version=__version__, current_time=current_time, system_info=system_info,
                                                    test_suite_json=test_suite_json)

    f = open(os.path.join(report_dir, "index.html"), mode="w", encoding="utf-8")
    try:
        f.write(index_page_content)
        pconsole.write_line("html report is generated at %s" % os.path.abspath(report_dir))
    except Exception as e:
        pconsole.write_line("Failed to generate html report.\n%s" % traceback.format_exc())
    finally:
        f.close()


def _get_test_suite_dict(test_suite):
    test_suite_dict = {
        "name": escape_html(test_suite.name),
        "fullName": escape_html(test_suite.full_name),
        "type": "suite",
        "testModules": _get_test_module_dicts(test_suite.test_classes),
        "startTime": str(test_suite.start_time),
        "endTime": str(test_suite.end_time),
        "elapsedTime": test_suite.elapsed_time,
        "total": test_suite.status_count[TestCaseCountItem.TOTAL],
        "passed": test_suite.status_count[TestCaseCountItem.PASSED],
        "failed": test_suite.status_count[TestCaseCountItem.FAILED],
        "skipped": test_suite.status_count[TestCaseCountItem.SKIPPED]
    }
    if not test_suite.before_suite.is_empty:
        test_suite_dict["beforeSuite"] = _get_test_fixture_dict(test_suite.before_suite)
    if not test_suite.after_suite.is_empty:
        test_suite_dict["afterSuite"] =_get_test_fixture_dict(test_suite.after_suite)
    return test_suite_dict


def _get_test_module_dicts(test_classes):
    root_test_module_dict = {
        "name": "root",
        "testModules": []
    }

    def get_or_new_module(modules, module_full_name):
        for module in modules:
            if module_full_name == module["fullName"]:
                return module
        new_module = {
            "name": module_full_name.split(".")[-1],
            "fullName": module_full_name,
            "type": "module",
            "testModules": [],
            "testClasses": [],
            "total": 0,
            "passed": 0,
            "failed": 0,
            "skipped": 0
        }
        modules.append(new_module)
        modules.sort(key=lambda m: m["name"])
        return new_module

    for test_class_dict in [_get_test_class_dict(test_class) for test_class in test_classes]:
        current_test_module_dict = root_test_module_dict
        splitted_full_name = test_class_dict["fullName"].split(".")[:-1]
        for i in range(len(splitted_full_name)):
            test_module_dict = get_or_new_module(current_test_module_dict["testModules"],
                                                 ".".join(splitted_full_name[:i + 1]))
            test_module_dict["total"] += test_class_dict["total"]
            test_module_dict["passed"] += test_class_dict["passed"]
            test_module_dict["failed"] += test_class_dict["failed"]
            test_module_dict["skipped"] += test_class_dict["skipped"]
            current_test_module_dict = test_module_dict
        current_test_module_dict["testClasses"].append(test_class_dict)
        current_test_module_dict["testClasses"].sort(key=lambda c: c["name"])
    return root_test_module_dict["testModules"]


def _get_test_class_dict(test_class):
    test_class_dict = {
        "name": test_class.name,
        "fullName": test_class.full_name,
        "type": "class",
        "runMode": test_class.run_mode,
        "runGroup": test_class.run_group,
        "description": test_class.description,
        "startTime": str(test_class.start_time),
        "endTime": str(test_class.end_time),
        "elapsedTime": test_class.elapsed_time,
        "total": test_class.status_count[TestCaseCountItem.TOTAL],
        "passed": test_class.status_count[TestCaseCountItem.PASSED],
        "failed": test_class.status_count[TestCaseCountItem.FAILED],
        "skipped": test_class.status_count[TestCaseCountItem.SKIPPED]
    }
    if test_class.is_group_feature_used:
        test_class_dict["testGroups"] = sorted([_get_test_group_dict(test_group) for test_group in test_class.test_groups], key=lambda g: g["name"])
    else:
        test_class_dict["testCases"] = sorted([_get_test_case_dict(test_case) for test_case in test_class.test_cases], key=lambda c: c["name"])

    if not test_class.before_class.is_empty:
        test_class_dict["beforeClass"] = _get_test_fixture_dict(test_class.before_class)
    if not test_class.after_class.is_empty:
        test_class_dict["afterClass"] =_get_test_fixture_dict(test_class.after_class)
    return test_class_dict


def _get_test_group_dict(test_group):
    test_group_dict = {
        "name": escape_html(test_group.name),
        "fullName": escape_html(test_group.full_name),
        "type": "group",
        "testCases": sorted([_get_test_case_dict(test_case) for test_case in test_group.test_cases], key=lambda c: c["name"]),
        "startTime": str(test_group.start_time),
        "endTime": str(test_group.end_time),
        "elapsedTime": test_group.elapsed_time,
        "total": test_group.status_count[TestCaseCountItem.TOTAL],
        "passed": test_group.status_count[TestCaseCountItem.PASSED],
        "failed": test_group.status_count[TestCaseCountItem.FAILED],
        "skipped": test_group.status_count[TestCaseCountItem.SKIPPED]
    }
    if not test_group.before_group.is_empty:
        test_group_dict["beforeGroup"] = _get_test_fixture_dict(test_group.before_group)
    if not test_group.after_group.is_empty:
        test_group_dict["afterGroup"] = _get_test_fixture_dict(test_group.after_group)
    return test_group_dict


def _get_test_case_dict(test_case):
    test_case_dict = {
        "name": escape_html(test_case.name),
        "fullName": escape_html(test_case.full_name),
        "type": "case",
        "startTime": str(test_case.start_time),
        "endTime": str(test_case.end_time),
        "elapsedTime": test_case.elapsed_time,
        "status": test_case.status,
        "tags": test_case.tags,
        "group": test_case.group,
        "description": test_case.description,
        "test": _get_test_fixture_dict(test_case.test),
    }
    if not test_case.before_method.is_empty:
        test_case_dict["beforeMethod"] = _get_test_fixture_dict(test_case.before_method)
    if not test_case.after_method.is_empty:
        test_case_dict["afterMethod"] = _get_test_fixture_dict(test_case.after_method)
    return test_case_dict


def _get_test_fixture_dict(test_fixture):
    test_fixture_dict = {
        "name": escape_html(test_fixture.name),
        "fullName": escape_html(test_fixture.full_name),
        "type": "fixture",
        "status": test_fixture.status,
        "fixtureType": test_fixture.fixture_type,
        "startTime": str(test_fixture.start_time),
        "endTime": str(test_fixture.end_time),
        "elapsedTime": test_fixture.elapsed_time,
        "logs": escape_html(test_fixture.logs),
        "description": test_fixture.description
    }
    return test_fixture_dict



from copy import copy
from functools import cmp_to_key
import threading
import traceback
import time

from datetime import datetime

from .enumeration import TestFixtureStatus, TestClassRunMode, TestCaseStatus
from .plogger import preporter, pconsole, pconsole_err
from .testsuite import AfterSuite, BeforeSuite, AfterClass, BeforeClass, BeforeGroup, AfterGroup, AfterMethod, \
    BeforeMethod, Test
from .plistener import test_listeners
from .utils import call_function, kill_thread, format_thread_stack


class TestExecutor(threading.Thread):
    def __init__(self, parent_test_executor):
        threading.Thread.__init__(self)
        self.parent_test_executor = parent_test_executor
        self.__properties = {}
        if self.parent_test_executor:
            for key, value in self.parent_test_executor.get_properties().items():
                if isinstance(value, (list, tuple, set, dict)):
                    self.__properties[key] = copy(value)
                else:
                    self.__properties[key] = value
        self.workers = 0
        self._lock = threading.RLock()

    def start_and_join(self):
        self.start()
        self.join()

    def update_properties(self, properties):
        self.__properties.update(properties)

    def clear_properties(self):
        self.__properties.clear()

    def get_property(self, key):
        try:
            return self.__properties[key]
        except KeyError:
            return None

    def get_properties(self):
        return self.__properties

    def allocate_worker(self, child_test_executor):
        with self._lock:
            if self.workers > 0:
                self.workers -= 1
                child_test_executor.workers += 1
                return True
            else:
                return False

    def acquire_worker(self):
        if self.parent_test_executor:
            with self._lock:
                if self.parent_test_executor.allocate_worker(self):
                    return True
                else:
                    if self.parent_test_executor.acquire_worker():
                        return self.parent_test_executor.allocate_worker(self)
                    else:
                        return False
        else:
            with self._lock:
                return self.workers > 0

    def release_worker(self):
        if self.parent_test_executor:
            with self.parent_test_executor._lock:
                self.parent_test_executor.workers += self.workers
                self.workers = 0
        else:
            pass


class TestSuiteExecutor(TestExecutor):
    def __init__(self, test_suite, workers):
        TestExecutor.__init__(self, None)
        self.test_suite = test_suite
        self.workers = workers

    def run(self):
        before_suite_executor = TestFixtureExecutor(self, self.test_suite.before_suite)
        before_suite_executor.acquire_worker()
        test_listeners.on_test_suite_start(self.test_suite)
        self.test_suite.start_time = datetime.now()
        before_suite_executor.start_and_join()
        before_suite_executor.release_worker()

        test_class_run_group_executors = []

        for test_class_run_group in self.test_suite.test_class_run_groups:
            test_class_run_group_executor = TestClassRunGroupExecutor(self, test_class_run_group)
            test_class_run_group_executors.append(test_class_run_group_executor)
            test_class_run_group_executor.start()

        for executor in test_class_run_group_executors:
            executor.join()

        after_suite_executor = TestFixtureExecutor(self, self.test_suite.after_suite)
        after_suite_executor.acquire_worker()
        after_suite_executor.start_and_join()
        after_suite_executor.release_worker()
        self.test_suite.end_time = datetime.now()
        test_listeners.on_test_suite_finish(self.test_suite)

        self.release_worker()

class TestClassRunGroupExecutor(TestExecutor):
    def __init__(self, test_suite_executor, test_class_run_group):
        TestExecutor.__init__(self, test_suite_executor)
        self.test_class_run_group = test_class_run_group

    def run(self):
        for test_class in self.test_class_run_group:
            TestClassExecutor(self, test_class).start_and_join()

        self.release_worker()

class TestClassExecutor(TestExecutor):
    def __init__(self, test_class_run_group_executor, test_class):
        TestExecutor.__init__(self, test_class_run_group_executor)
        self.test_class = test_class

    def run(self):
        before_class_executor = TestFixtureExecutor(self, self.test_class.before_class)
        before_class_executor.acquire_worker()
        test_listeners.on_test_class_start(self.test_class)
        self.test_class.start_time = datetime.now()
        before_class_executor.start_and_join()
        before_class_executor.release_worker()

        if self.test_class.run_mode == TestClassRunMode.SingleLine:
            for test_group in self.test_class.test_groups:
                TestGroupExecutor(self, test_group).start_and_join()
        else:
            test_group_executors = []

            for test_group in self.test_class.test_groups:
                test_group_executor = TestGroupExecutor(self, test_group)
                test_group_executors.append(test_group_executor)
                test_group_executor.start()

            for executor in test_group_executors:
                executor.join()

        after_class_executor = TestFixtureExecutor(self, self.test_class.after_class)
        after_class_executor.acquire_worker()
        after_class_executor.start_and_join()
        after_class_executor.release_worker()
        self.test_class.end_time = datetime.now()
        test_listeners.on_test_class_finish(self.test_class)

        self.release_worker()


class TestGroupExecutor(TestExecutor):
    def __init__(self, test_class_executor, test_group):
        TestExecutor.__init__(self, test_class_executor)
        self.test_group = test_group

    def run(self):
        before_group_executor = TestFixtureExecutor(self, self.test_group.before_group)
        before_group_executor.acquire_worker()
        test_listeners.on_test_group_start(self.test_group)
        self.test_group.start_time = datetime.now()
        before_group_executor.start_and_join()
        before_group_executor.release_worker()

        if self.test_group.test_class.run_mode == TestClassRunMode.SingleLine:
            for test_case in self.test_group.test_cases:
                TestCaseExecutor(self, test_case).start_and_join()
        else:
            test_case_executors = []

            for test_case in self.test_group.test_cases:
                test_case_executor = TestCaseExecutor(self, test_case)
                test_case_executors.append(test_case_executor)
                test_case_executor.start()

            for executor in test_case_executors:
                executor.join()

        after_group_executor = TestFixtureExecutor(self, self.test_group.after_group)
        after_group_executor.acquire_worker()
        after_group_executor.start_and_join()
        after_group_executor.release_worker()
        self.test_group.end_time = datetime.now()
        test_listeners.on_test_group_finish(self.test_group)

        self.release_worker()


class TestCaseExecutor(TestExecutor):
    def __init__(self, test_group_executor, test_case):
        TestExecutor.__init__(self, test_group_executor)
        self.test_case = test_case

    def run(self):
        before_method_executor = TestFixtureExecutor(self, self.test_case.before_method)
        before_method_executor.acquire_worker()
        test_listeners.on_test_case_start(self.test_case)
        self.test_case.start_time = datetime.now()
        before_method_executor.start_and_join()
        before_method_executor.release_worker()

        test_executor = TestFixtureExecutor(self, self.test_case.test)
        test_executor.acquire_worker()
        test_executor.start_and_join()

        logger_filler = "-" * (100 - len(self.test_case.full_name) - 6)
        if self.test_case.status == TestCaseStatus.PASSED:
            pconsole.write_line("%s%s|PASS|" % (self.test_case.full_name, logger_filler))
        elif self.test_case.status == TestCaseStatus.FAILED:
            pconsole.write_line("%s%s|FAIL|" % (self.test_case.full_name, logger_filler))
        elif self.test_case.status == TestCaseStatus.SKIPPED:
            pconsole.write_line("%s%s|SKIP|" % (self.test_case.full_name, logger_filler))

        test_executor.release_worker()

        after_method_executor =TestFixtureExecutor(self, self.test_case.after_method)
        after_method_executor.acquire_worker()
        after_method_executor.start_and_join()
        after_method_executor.release_worker()
        self.test_case.end_time = datetime.now()
        test_listeners.on_test_case_finish(self.test_case)

        self.release_worker()


class TestFixtureExecutor(TestExecutor):
    def __init__(self, parent_test_executor, test_fixture):
        TestExecutor.__init__(self, parent_test_executor)
        self.test_fixture = test_fixture

    def run_test_fixture(self):
        if self.test_fixture.is_empty:
            self.test_fixture.status = TestFixtureStatus.PASSED
            return

        self.test_fixture.status = TestFixtureStatus.RUNNING
        test_fixture_sub_executor = TestFixtureSubExecutor(self)
        test_fixture_sub_executor.start()
        if self.test_fixture.timeout > 0:
            test_fixture_sub_executor.join(self.test_fixture.timeout)
            if test_fixture_sub_executor.isAlive():
                stack_trace = format_thread_stack(test_fixture_sub_executor)
                try:
                    kill_thread(test_fixture_sub_executor)
                except Exception as e:
                    pconsole_err.write_line(e)
                from .plogger import preporter
                self.test_fixture.status = TestFixtureStatus.FAILED
                self.test_fixture.failure_message = "Timed out executing this test fixture in %s seconds." % self.test_fixture.timeout
                self.test_fixture.failure_type = "TimeoutException"
                self.test_fixture.stack_trace = stack_trace
                preporter.error("Failed with following message:\n%s\n%s" % (self.test_fixture.failure_message, self.test_fixture.stack_trace), True)
        else:
            test_fixture_sub_executor.join()

    def skip_test_fixture(self, caused_test_fixture):
        if self.test_fixture.is_empty:
            self.test_fixture.status = TestFixtureStatus.SKIPPED
            return

        from .plogger import preporter
        self.test_fixture.status = TestFixtureStatus.SKIPPED
        self.test_fixture.skip_message = "@%s failed, so skipped." % caused_test_fixture.fixture_type
        preporter.warn("@%s failed, so skipped." % caused_test_fixture.fixture_type)

    def run(self):
        self.test_fixture.start_time = datetime.now()
        self.update_properties({"running_test_fixture": self.test_fixture})

        failed_setup_fixture = self.test_fixture.context.get_failed_setup_fixture()
        if not failed_setup_fixture:
            self.run_test_fixture()
        elif isinstance(self.test_fixture, AfterSuite) and isinstance(failed_setup_fixture, BeforeSuite) and self.test_fixture.always_run:
            self.run_test_fixture()
        elif isinstance(self.test_fixture, AfterClass) and isinstance(failed_setup_fixture, BeforeClass) and self.test_fixture.always_run:
            self.run_test_fixture()
        elif isinstance(self.test_fixture, AfterGroup) and isinstance(failed_setup_fixture, BeforeGroup) and self.test_fixture.always_run:
            self.run_test_fixture()
        elif isinstance(self.test_fixture, AfterMethod) and isinstance(failed_setup_fixture, BeforeMethod) and self.test_fixture.always_run:
            self.run_test_fixture()
        else:
            self.skip_test_fixture(failed_setup_fixture)

        # spread before's attributes
        if not self.test_fixture.is_empty:
            if isinstance(self.test_fixture, BeforeSuite):
                before_suite_dict = self.test_fixture.test_fixture_ref.__self__.__dict__
                for test_class in self.test_fixture.test_suite.test_classes:
                    test_class.test_class_ref.__dict__.update(before_suite_dict)
                    for test_group in test_class.test_groups:
                        test_group.test_class_ref.__dict__.update(before_suite_dict)
                        for test_case in test_group.test_cases:
                            test_case.test_case_ref.__self__.__dict__.update(before_suite_dict)
            elif isinstance(self.test_fixture, BeforeClass):
                before_class_dict = self.test_fixture.test_fixture_ref.__self__.__dict__
                for test_group in self.test_fixture.test_class.test_groups:
                    test_group.test_class_ref.__dict__.update(before_class_dict)
                    for test_case in test_group.test_cases:
                        test_case.test_case_ref.__self__.__dict__.update(before_class_dict)
            elif isinstance(self.test_fixture, BeforeGroup):
                before_group_dict = self.test_fixture.test_fixture_ref.__self__.__dict__
                for test_case in self.test_fixture.test_group.test_cases:
                    test_case.test_case_ref.__self__.__dict__.update(before_group_dict)

        self.update_properties({"running_test_fixture": None})
        self.test_fixture.end_time = datetime.now()

    def acquire_worker(self):
        while True:
            if TestExecutor.acquire_worker(self):
                return
            else:
                time.sleep(1)


class TestFixtureSubExecutor(TestExecutor):
    def __init__(self, test_fixture_executor):
        TestExecutor.__init__(self, test_fixture_executor)
        self.test_fixture = test_fixture_executor.test_fixture
        self.setDaemon(True)

    def run_test(self):
        if self.test_fixture.expected_exceptions:
            expected_exceptions = self.test_fixture.expected_exceptions
            expected_exceptions_names = str(["%s.%s" % (e.__module__, e.__name__) for e in expected_exceptions.keys()])
            try:
                params = self.test_fixture.parameters or []
                call_function(self.test_fixture.test_fixture_ref, *params)
            except Exception as e:
                exception = e.__class__
                exception_name = "%s.%s" % (exception.__module__, exception.__name__)
                matched_exceptions = [expected_exception for expected_exception in expected_exceptions.keys() if
                                      issubclass(exception, expected_exception)]
                if matched_exceptions:
                    def cmp_matched_exception(exception_a, exception_b):
                        return -1 if issubclass(exception_a, exception_b) else 1

                    matched_exception = sorted(matched_exceptions, key=cmp_to_key(cmp_matched_exception))[0]
                    if not expected_exceptions[matched_exception] or expected_exceptions[matched_exception].search(str(e)):
                        self.test_fixture.status = TestFixtureStatus.PASSED
                    else:
                        self.test_fixture.status = TestFixtureStatus.FAILED
                        self.test_fixture.failure_message = "The exception <%s> was thrown with the wrong message: Expected message regex: <%s>, Actual message: <%s>." \
                                                            % (exception_name, expected_exceptions[matched_exception].pattern, str(e))
                        self.test_fixture.failure_type = "WrongExceptionMessageError"
                        self.test_fixture.stack_trace = traceback.format_exc()
                        preporter.error("Failed with following message:\n%s\n%s" % (self.test_fixture.failure_message, self.test_fixture.stack_trace), True)
                else:
                    self.test_fixture.status = TestFixtureStatus.FAILED
                    self.test_fixture.failure_message = "Expected exception: one of %s, Actual exception: <%s>." \
                                                        % (expected_exceptions_names, exception_name)
                    self.test_fixture.failure_type = "WrongExceptionThrownError"
                    self.test_fixture.stack_trace = traceback.format_exc()
                    preporter.error("Failed with following message:\n%s\n%s" % (self.test_fixture.failure_message, self.test_fixture.stack_trace), True)
            else:
                self.test_fixture.status = TestFixtureStatus.FAILED
                self.test_fixture.failure_message = "Expected exception: one of %s, Actual: NO exception was thrown." \
                                                    % expected_exceptions_names
                self.test_fixture.failure_type = "NoExceptionThrownError"
                self.test_fixture.stack_trace = self.test_fixture.failure_message
                preporter.error("Failed with following message:\n%s" % self.test_fixture.failure_message, True)
        else:
            try:
                params = self.test_fixture.parameters or []
                call_function(self.test_fixture.test_fixture_ref, *params)
            except Exception as e:
                self.test_fixture.status = TestFixtureStatus.FAILED
                self.test_fixture.failure_message = str(e).strip() or "\n".join([str(arg) for arg in e.args])
                self.test_fixture.failure_type = "%s.%s" % (e.__class__.__module__, e.__class__.__name__)
                self.test_fixture.stack_trace = traceback.format_exc()
                preporter.error("Failed with following message:\n%s" % self.test_fixture.stack_trace, True)
            else:
                self.test_fixture.status = TestFixtureStatus.PASSED

    def run_test_configuration(self):
        try:
            params = {1: [], 2: [self.test_fixture.context]}[self.test_fixture.arguments_count]
            call_function(self.test_fixture.test_fixture_ref, *params)
        except Exception as e:
            self.test_fixture.status = TestFixtureStatus.FAILED
            self.test_fixture.failure_message = str(e).strip() or "\n".join([str(arg) for arg in e.args])
            self.test_fixture.failure_type = "%s.%s" % (e.__class__.__module__, e.__class__.__name__)
            self.test_fixture.stack_trace = traceback.format_exc()
            preporter.error("Failed with following message:\n%s" % self.test_fixture.stack_trace, True)
        else:
            self.test_fixture.status = TestFixtureStatus.PASSED

    def run(self):
        if isinstance(self.test_fixture, Test):
            self.run_test()
        else:
            self.run_test_configuration()


def current_executor():
    return threading.currentThread()



import importlib
import os
import shlex
import traceback
from codecs import open
from xml.dom import minidom

from .utils import make_dirs, remove_tree, StringTypes


def get_rerun_targets(xml_file):
    test_targets = []
    doc = minidom.parse(xml_file)
    if doc.documentElement.nodeName == "testsuites":
        root = doc.documentElement
    else:
        root = doc
    for test_suite_node in root.getElementsByTagName("testsuite"):
        for test_case_node in test_suite_node.getElementsByTagName("testcase"):
            if test_case_node.getElementsByTagName("failure") or test_case_node.getElementsByTagName("skipped"):
                test_target = "%s.%s" % (test_case_node.getAttribute("classname"), test_case_node.getAttribute("name"))
                test_targets.append(test_target)
    return test_targets


def merge_xunit_xmls(xml_files, to_file):
    from .plogger import pconsole
    from .testsuite import default_test_suite

    pconsole.write_line("Start to merge xunit result xmls...")

    test_case_results = {}

    for xml_file in xml_files:
        doc = minidom.parse(xml_file)
        if doc.documentElement.nodeName == "testsuites":
            root = doc.documentElement
        else:
            root = doc
        for test_suite_node in root.getElementsByTagName("testsuite"):
            for test_case_node in test_suite_node.getElementsByTagName("testcase"):
                test_case_name = "%s.%s" % (test_case_node.getAttribute("classname"), test_case_node.getAttribute("name"))
                test_case_status = 0 # passed
                if test_case_node.getElementsByTagName("failure"):
                    test_case_status = 1 # failed
                elif test_case_node.getElementsByTagName("skipped"):
                    test_case_status = 2 # skipped

                if test_case_name not in test_case_results or test_case_status < test_case_results[test_case_name]["status"]:
                    test_case_results[test_case_name] = {"status": test_case_status, "node": test_case_node}

    doc = minidom.Document()
    test_suite_ele = doc.createElement("testsuite")
    doc.appendChild(test_suite_ele)
    test_suite_ele.setAttribute("name", default_test_suite.name)
    test_suite_ele.setAttribute("tests", str(len(test_case_results)))
    test_suite_ele.setAttribute("failures", str(len([result for result in test_case_results.values() if result["status"] == 1])))
    test_suite_ele.setAttribute("skips", str(len([result for result in test_case_results.values() if result["status"] == 2])))
    test_suite_ele.setAttribute("errors", "0")

    for test_case_result in test_case_results.values():
        test_suite_ele.appendChild(test_case_result["node"])

    if os.path.exists(to_file):
        pconsole.write_line("Cleaning old merged xunit result xml...")
        os.remove(to_file)
    else:
        make_dirs(os.path.dirname(to_file))

    f = open(to_file, mode="w", encoding="utf-8")
    try:
        doc.writexml(f, "\t", "\t", "\n", "utf-8")
        pconsole.write_line("Merged xunit xml is generated at %s" % to_file)
    except Exception:
        pconsole.write_line("Failed to generate merged xunit xml.\n%s" % traceback.format_exc())
    finally:
        f.close()


def main(args=None):
    # fix encoding issue in python 2.x
    import sys
    try:
        reload(sys)
        sys.setdefaultencoding('utf-8')
    except NameError:
        pass

    from . import config
    # load arguments
    if args is None:
        args = sys.argv[1:]
    elif not isinstance(args, (tuple, list)):
        if not isinstance(args, StringTypes):
            sys.stderr.write("ERROR: args <%s> is not a string or argument list." % args)
            return
        args = shlex.split(args)
    config.load(args)

    # merge xunit result xmls
    xunit_xmls = config.get_option("merge_xunit_xmls")
    if xunit_xmls is not None:
        merge_xunit_xmls(xunit_xmls, config.get_option("to"))
        return

    # run test
    from .testfilter import TestFilterGroup, TestIncludeTagsFilter, TestExcludeTagsFilter, TestIncludeGroupsFilter
    from . import testexecutor, reporter, plistener
    from .testfinder import TestFinder
    from .testsuite import default_test_suite
    from .plogger import pconsole
    from .enumeration import TestCaseCountItem

    pconsole.write_line("Starting ptest...")

    # add workspace to python path
    workspace = config.get_option("workspace")
    sys.path.insert(0, workspace)
    pconsole.write_line("Workspace:")
    pconsole.write_line(" %s" % workspace)

    # add python_paths to python path
    python_paths = config.get_option("python_paths")
    if python_paths is not None:
        pconsole.write_line("Python paths:")
        for python_path in python_paths:
            sys.path.append(python_path)
            pconsole.write_line(" %s" % python_path)

    # test filter group
    test_filter_group = TestFilterGroup()

    include_tags = config.get_option("include_tags")
    if include_tags is not None:
        test_filter_group.append_filter(TestIncludeTagsFilter(include_tags))

    exclude_tags = config.get_option("exclude_tags")
    if exclude_tags is not None:
        test_filter_group.append_filter(TestExcludeTagsFilter(exclude_tags))

    include_groups = config.get_option("include_groups")
    if include_groups is not None:
        test_filter_group.append_filter(TestIncludeGroupsFilter(include_groups))

    filter_path = config.get_option("test_filter")
    if filter_path is not None:
        splitted_filter_path = filter_path.split(".")
        filter_module = importlib.import_module(".".join(splitted_filter_path[:-1]))
        filter_class = getattr(filter_module, splitted_filter_path[-1])
        test_filter_group.append_filter(filter_class())

    if test_filter_group:
        pconsole.write_line("Test filters:")
        for test_filter in test_filter_group:
            pconsole.write_line(" %s" % test_filter)

    # get test targets
    test_targets = config.get_option("test_targets")
    if test_targets is not None:
        pconsole.write_line("Test targets:")
        for test_target in test_targets:
            test_finder = TestFinder(test_target, test_filter_group, default_test_suite)
            test_finder.find_tests()
            if test_finder.repeated_test_count:
                pconsole.write_line(" %s (%s tests found, %s repeated)" % (test_target, test_finder.found_test_count, test_finder.repeated_test_count))
            else:
                pconsole.write_line(" %s (%s tests found)" % (test_target, test_finder.found_test_count))
    else:
        # rerun failed/skipped test cases
        pconsole.write_line("Run failed/skipped tests in xunit xml:")
        xunit_xml = config.get_option("run_failed")
        test_targets = get_rerun_targets(xunit_xml)
        found_test_count = 0
        for test_target in test_targets:
            test_finder = TestFinder(test_target, test_filter_group, default_test_suite)
            test_finder.find_tests()
            found_test_count += test_finder.found_test_count
        pconsole.write_line(" %s (%s tests found)" % (xunit_xml, found_test_count))

    # add test listeners
    listener_paths = config.get_option("test_listeners")
    if listener_paths is not None:
        pconsole.write_line("Test listeners:")
        for listener_path in listener_paths:
            pconsole.write_line(" %s" % listener_path)
            splitted_listener_path = listener_path.split(".")
            listener_module = importlib.import_module(".".join(splitted_listener_path[:-1]))
            listener_class = getattr(listener_module, splitted_listener_path[-1])
            plistener.test_listeners.append(listener_class())

    # init test suite
    default_test_suite.init()
    test_cases = default_test_suite.test_cases

    # exit if no tests found
    if len(test_cases) == 0:
        pconsole.write_line("=" * 100)
        pconsole.write_line("No tests found. Please check your command line options.")
        return

    # add webdriver instance to test executor to support capturing screenshot for webdriver
    try:
        from selenium.webdriver.remote.webdriver import WebDriver
    except ImportError as e:
        pass
    else:
        def add_web_driver(executor, web_driver):
            web_drivers = executor.get_property("web_drivers")
            if web_drivers is None:
                web_drivers = []
                executor.update_properties({"web_drivers": web_drivers})
            web_drivers.append(web_driver)
        def new_start_client(self):
            try:
                current_executor = testexecutor.current_executor()
                add_web_driver(current_executor, self)
                add_web_driver(current_executor.parent_test_executor, self)
                add_web_driver(current_executor.parent_test_executor.parent_test_executor, self)
            except AttributeError as e:
                pass

        def remove_web_driver(executor, web_driver):
            web_drivers = executor.get_property("web_drivers")
            if web_drivers:
                web_drivers.remove(web_driver)
        def new_stop_client(self):
            try:
                current_executor = testexecutor.current_executor()
                remove_web_driver(current_executor, self)
                remove_web_driver(current_executor.parent_test_executor, self)
                remove_web_driver(current_executor.parent_test_executor.parent_test_executor, self)
            except AttributeError as e:
                pass
        WebDriver.start_client = new_start_client
        WebDriver.stop_client = new_stop_client

    # print test names
    pconsole.write_line("=" * 100)
    pconsole.write_line("Start to run following %s tests:" % len(test_cases))
    pconsole.write_line("-" * 30)
    for test_case in test_cases:
        pconsole.write_line(" %s" % test_case.full_name)
    pconsole.write_line("=" * 100)

    # clean and create temp dir
    temp_dir = config.get_option("temp")
    if os.path.exists(temp_dir):
        remove_tree(temp_dir, remove_root=False)
    else:
        make_dirs(temp_dir)

    # run test cases
    testexecutor.TestSuiteExecutor(default_test_suite, int(config.get_option("test_executor_number"))).start_and_join()

    # log the test results
    status_count = default_test_suite.status_count
    pconsole.write_line("")
    pconsole.write_line("=" * 100)
    pconsole.write_line("Test finished in %.2fs." % default_test_suite.elapsed_time)
    pconsole.write_line("Total: %s, passed: %s, failed: %s, skipped: %s. Pass rate: %.1f%%." % (
        status_count[TestCaseCountItem.TOTAL], status_count[TestCaseCountItem.PASSED],
        status_count[TestCaseCountItem.FAILED], status_count[TestCaseCountItem.SKIPPED], default_test_suite.pass_rate))

    # generate the test report
    pconsole.write_line("")
    pconsole.write_line("=" * 100)
    reporter.generate_xunit_xml(config.get_option("xunit_xml"))
    reporter.generate_html_report(config.get_option("report_dir"))

    # clean temp dir
    remove_tree(temp_dir)



import importlib
import pkgutil
import re

from .enumeration import PDecoratorType
from .utils import mock_func


class TestFinder:
    def __init__(self, test_target, test_filter_group, target_test_suite):
        self.test_target = test_target
        self.test_filter_group = test_filter_group
        self.target_test_suite = target_test_suite
        self.found_test_count = 0
        self.repeated_test_count = 0
        # test class / test case name filter
        self.test_class_name = None
        self.test_name = None
        self.is_parametric_test = None

    def find_tests(self):
        splitted_test_target = self.test_target.split(".")
        module_ref = None
        module_name_len = 0
        for i in range(len(splitted_test_target)):
            try:
                module_ref = importlib.import_module(".".join(splitted_test_target[:i + 1]))
                module_name_len = i + 1
            except ImportError as e:
                if splitted_test_target[i] in str(e):
                    break
                raise

        if module_ref is None:
            raise ImportError("Test target <%s> is invalid.\nNo module named <%s>."% (self.test_target, splitted_test_target[0]))

        test_target_len = len(splitted_test_target)
        if module_name_len == test_target_len:
            if hasattr(module_ref, "__path__"):
                # test target is package
                self.find_tests_in_package(module_ref)
            else:
                # test target is module
                self.find_tests_in_module(module_ref)
        elif module_name_len == test_target_len - 1:
            # test target is class
            self.test_class_name = splitted_test_target[-1]
            self.find_tests_in_module(module_ref)
        elif module_name_len == test_target_len - 2:
            # test target is method
            self.test_class_name = splitted_test_target[-2]
            self.test_name = splitted_test_target[-1]
            self.is_parametric_test = re.search(r"^.*#.*$", self.test_name)
            self.find_tests_in_module(module_ref)
        else:
            raise ImportError("Test target <%s> is probably invalid.\nModule <%s> exists but module <%s> doesn't."% (
                self.test_target, ".".join(splitted_test_target[:module_name_len]), ".".join(splitted_test_target[:module_name_len + 1])))

    def find_tests_in_package(self, package_ref):
        for _, name, is_package in pkgutil.iter_modules(package_ref.__path__):
            if is_package:
                self.find_tests_in_package(importlib.import_module(package_ref.__name__ + "." + name))
            else:
                self.find_tests_in_module(importlib.import_module(package_ref.__name__ + "." + name))

    def find_tests_in_module(self, module_ref):
        for module_element in dir(module_ref):
            test_class_cls = getattr(module_ref, module_element)
            if hasattr(test_class_cls, "__pd_type__") and test_class_cls.__pd_type__ == PDecoratorType.TestClass \
                    and hasattr(test_class_cls, "__enabled__") and test_class_cls.__enabled__ \
                    and test_class_cls.__module__ == module_ref.__name__ \
                    and (not self.test_class_name or self.test_class_name == test_class_cls.__name__):
                self.find_tests_in_class(test_class_cls)

    def find_tests_in_class(self, test_class_cls):
        for class_element in dir(test_class_cls):
            test_func = getattr(test_class_cls, class_element)
            if hasattr(test_func, "__pd_type__") and test_func.__pd_type__ == PDecoratorType.Test \
                    and hasattr(test_func, "__enabled__") and test_func.__enabled__:
                if not self.test_name:
                    for func in unzip_func(test_class_cls, test_func):
                        if self.test_filter_group.filter(func):
                            self.__add_test(test_class_cls, func)
                else:
                    if self.is_parametric_test and test_func.__data_provider__:
                        if re.search(r"^%s#.*$" % test_func.__name__, self.test_name):
                            for func in unzip_func(test_class_cls, test_func):
                                if self.test_name == func.__name__ and self.test_filter_group.filter(func):
                                    self.__add_test(test_class_cls, func)
                    elif self.test_name == test_func.__name__:
                        for func in unzip_func(test_class_cls, test_func):
                            if self.test_filter_group.filter(func):
                                self.__add_test(test_class_cls, func)

    def __add_test(self, test_class_cls, func):
        self.found_test_count += 1
        if not self.target_test_suite.add_test_case(test_class_cls, func):
            self.repeated_test_count += 1


def unzip_func(test_class_cls, test_func):
    if not test_func.__funcs__: # zipped
        name_map = {}
        for index, data in enumerate(test_func.__data_provider__):
            if isinstance(data, (list, tuple)):
                parameters_number = len(data)
                parameters = data
            else:
                parameters_number = 1
                parameters = [data]
            if parameters_number == test_func.__arguments_count__ - 1:
                mock = mock_func(test_func)
                mock_name = ("%s#%s" % (test_func.__name__, test_func.__data_name__(index, parameters))) \
                    .replace(".", "_").replace(",", "_").replace(" ", "_")
                if mock_name in name_map:
                    name_map[mock_name] += 1
                    mock.__name__ = "%s(%s)" % (mock_name, name_map[mock_name] - 1)
                else:
                    name_map[mock_name] = 1
                    mock.__name__ = mock_name
                mock.__parameters__ = parameters
                mock.__data_index__ = index
                mock.__funcs__ = [mock]
                mock.__test_class__ = test_class_cls
                test_func.__funcs__.append(mock)
            else:
                raise TypeError("The data provider is trying to pass %s extra arguments but %s.%s() takes %s."
                                % (parameters_number, test_class_cls.__name__, test_func.__name__, test_func.__arguments_count__ - 1))
    elif not test_func.__data_provider__: # normal
        test_func.__funcs__[0].__test_class__ = test_class_cls
        if test_func.__arguments_count__ != 1:
            raise TypeError("Since data provider is not specified, %s.%s() cannot be declared with %s parameters. Please declare with only 1 parameter (self)."
                            % (test_class_cls.__name__, test_func.__name__, test_func.__arguments_count__))
    return test_func.__funcs__


import inspect
import os
import re

from .enumeration import PDecoratorType, TestClassRunMode
from .utils import urljoin, unquote, pathname2url, StringTypes


def TestClass(enabled=True, run_mode="singleline", run_group=None, description="", **custom_args):
    """
        The TestClass decorator, it is used to mark a class as TestClass.

    :param enabled: enable or disable this test class.
    :param run_mode: the run mode of all the test cases in this test class. If set to "parallel", all the test cases can be run by multiple threads. If set to "singleline", all the test cases will be only run by one thread.
    :param run_group: the run group of this test class. If run group is specified, all the test classes in the same run group will be run one by one. If not, this test class will be belong to it own run group.
    :param description: the description of this test class.
    :param custom_args: the custom arguments of this test class.
    """

    def tracer(cls):
        cls.__pd_type__ = PDecoratorType.TestClass
        cls.__enabled__ = enabled
        if run_mode.lower() in [TestClassRunMode.SingleLine, TestClassRunMode.Parallel]:
            cls.__run_mode__ = run_mode.lower()
        else:
            raise ValueError("Run mode <%s> is not supported. Please use <%s> or <%s>." % (
                run_mode, TestClassRunMode.Parallel, TestClassRunMode.SingleLine))
        cls.__run_group__ = None if run_group is None else str(run_group)
        cls.__description__ = description
        cls.__custom_args__ = custom_args
        return cls

    return tracer


def BeforeSuite(enabled=True, description="", timeout=0, **custom_args):
    """
        The BeforeSuite test fixture, it will be executed before test suite started.

    :param enabled: enable or disable this test fixture.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.BeforeSuite
        func.__enabled__ = enabled
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def BeforeClass(enabled=True, description="", timeout=0, **custom_args):
    """
        The BeforeClass test fixture, it will be executed before test class started.

    :param enabled: enable or disable this test fixture.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.BeforeClass
        func.__enabled__ = enabled
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def BeforeGroup(enabled=True, group="DEFAULT", description="", timeout=0, **custom_args):
    """
        The BeforeGroup test fixture, it will be executed before test group started.

    :param enabled: enable or disable this test fixture.
    :param group: the group that this test fixture belongs to.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.BeforeGroup
        func.__enabled__ = enabled
        func.__group__ = group
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def BeforeMethod(enabled=True, group="DEFAULT", description="", timeout=0, **custom_args):
    """
        The BeforeMethod test fixture, it will be executed before test started.

    :param enabled: enable or disable this test fixture.
    :param group: the group that this test fixture belongs to.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.BeforeMethod
        func.__enabled__ = enabled
        func.__group__ = group
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def Test(enabled=True, tags=[], expected_exceptions=None, data_provider=None, data_name=None, group="DEFAULT", description="", timeout=0, **custom_args):
    """
        The Test decorator, it is used to mark a test as Test.

    :param enabled: enable or disable this test.
    :param tags: the tags of this test. It can be string (separated by comma) or list or tuple.
    :param expected_exceptions: the expected exceptions of this test.
        If no exception or a different one is thrown, this test will be marked as failed.
        The possible values of this parameter are::
            Exception Class:
                expected_exceptions=AttributeError
            Exception Class list or tuple:
                expected_exceptions=[AttributeError, IndexError]
                expected_exceptions=(AttributeError, IndexError)
            Exception Class and regular expression of expected message dict:
                expected_exceptions={AttributeError: '.*object has no attribute.*'}
        Note: If you want to match the entire exception message, just include anchors in the regex pattern.
    :param data_provider: the data provider for this test, the data provider must be iterable.
        Following test will be run four times with the test data supplied from data provider.
            @Test(data_provider=[(1, 1, 2), (2, 3, 5), (4, 5, 9), (9, 9, 18)])
            def test_add(self, number1, number2, sum_):
                assert_that(number1 + number2).is_equal_to(sum_)
    :param data_name: the data name function of this test.
        Note: If no data_provider specified, data_name will be ignored.
        For example:
            @Test(data_provider=["foo", "bar"], data_name=lambda index, params: params[0])
            def test_something(self, name):
                assert_that(name).is_not_none()
            The test names are test_something#foo and test_something#bar.
    :param group: the group that this test belongs to.
    :param description: the description of this test.
    :param timeout: the timeout of this test (in seconds).
    :param custom_args: the custom arguments of this test.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.Test
        func.__enabled__ = enabled
        func.__group__ = group
        func.__description__ = description
        # deal with tags
        if not tags:
            func.__tags__ = []
        else:
            if isinstance(tags, StringTypes):
                tag_list = tags.split(",")
            elif isinstance(tags, (list, tuple)):
                tag_list = tags
            else:
                raise ValueError(
                    "Tags type %s is not supported. Please use string (separated by comma) or list or tuple." % type(tags))
            func.__tags__ = sorted([str(tag).strip() for tag in tag_list if str(tag).strip()])
        # deal with expected exceptions
        if not expected_exceptions:
            func.__expected_exceptions__ = None
        else:
            exceptions = {}
            if inspect.isclass(expected_exceptions):
                if issubclass(expected_exceptions, Exception):
                    exceptions[expected_exceptions] = None
                else:
                    raise ValueError("Expected exception should be a sub class of Exception.")
            elif isinstance(expected_exceptions, (tuple, list)):
                for exception in expected_exceptions:
                    if issubclass(exception, Exception):
                        exceptions[exception] = None
                    else:
                        raise ValueError("Expected exception should be a sub class of Exception.")
            elif isinstance(expected_exceptions, dict):
                for exception, message in expected_exceptions.items():
                    if issubclass(exception, Exception):
                        exceptions[exception] = re.compile(message)
                    else:
                        raise ValueError("Expected exception should be a sub class of Exception.")
            else:
                raise ValueError("Expected exceptions type %s is not supported. Please use class or list or tuple or dict."
                                 % type(expected_exceptions))
            func.__expected_exceptions__ = exceptions

        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test(func)
        # for data provider
        #                     normal    zipped    unzipped    mocked
        # __parameters__       None      None       None     not None
        # __data_index__       None      None       None     not None
        # __data_provider__    None    not None   not None   not None
        # __funcs__           [func]      []       [mocks]    [mock]
        func.__parameters__ = None
        func.__data_index__ = None
        func.__data_provider__ = None
        func.__funcs__ = [func]
        if data_provider:
            func.__data_provider__ = data_provider
            func.__funcs__ = []
            func.__data_name__ = data_name or (lambda index, params: index + 1)
            if len(inspect.getargspec(func.__data_name__)[0]) != 2:
                raise TypeError("Data name function must be declared with 2 parameters.")
        return func

    return handle_func


def AfterMethod(enabled=True, always_run=True, group="DEFAULT", description="", timeout=0, **custom_args):
    """
        The AfterMethod test fixture, it will be executed after test finished.

    :param enabled: enable or disable this test fixture.
    :param always_run: if set to true, this test fixture will be run even if the @BeforeMethod is failed. Otherwise, this test fixture will be skipped.
    :param group: the group that this test fixture belongs to.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.AfterMethod
        func.__enabled__ = enabled
        func.__always_run__ = always_run
        func.__group__ = group
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def AfterGroup(enabled=True, always_run=True, group="DEFAULT", description="", timeout=0, **custom_args):
    """
        The AfterGroup test fixture, it will be executed after test group finished.

    :param enabled: enable or disable this test fixture.
    :param always_run: if set to true, this test fixture will be run even if the @BeforeGroup is failed. Otherwise, this test fixture will be skipped.
    :param group: the group that this test fixture belongs to.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.AfterGroup
        func.__enabled__ = enabled
        func.__always_run__ = always_run
        func.__group__ = group
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def AfterClass(enabled=True, always_run=True, description="", timeout=0, **custom_args):
    """
        The AfterClass test fixture, it will be executed after test class finished.

    :param enabled: enable or disable this test fixture.
    :param always_run: if set to true, this test fixture will be run even if the @BeforeClass is failed. Otherwise, this test fixture will be skipped.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.AfterClass
        func.__enabled__ = enabled
        func.__always_run__ = always_run
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def AfterSuite(enabled=True, always_run=True, description="", timeout=0, **custom_args):
    """
        The AfterSuite test fixture, it will be executed after test suite finished.

    :param enabled: enable or disable this test fixture.
    :param always_run: if set to true, this test fixture will be run even if the @BeforeSuite is failed. Otherwise, this test fixture will be skipped.
    :param description: the description of this test fixture.
    :param timeout: the timeout of this test fixture (in seconds).
    :param custom_args: the custom arguments of this test fixture.
    """

    def handle_func(func):
        func.__pd_type__ = PDecoratorType.AfterSuite
        func.__enabled__ = enabled
        func.__always_run__ = always_run
        func.__description__ = description
        func.__timeout__ = timeout
        func.__custom_args__ = custom_args
        func.__location__ = __get_location(func)
        func.__arguments_count__ = __get_arguments_count_of_test_configuration(func)
        return func

    return handle_func


def __get_location(func):
    file_path = os.path.abspath(inspect.getfile(func))
    _, line_no = inspect.getsourcelines(func)
    return urljoin("file:", "%s:%s" % (unquote(pathname2url(file_path)), line_no))


def __get_arguments_count_of_test_configuration(func):
    arguments_count = len(inspect.getargspec(func)[0])
    if arguments_count not in [1, 2]:
        raise TypeError("%s() cannot be declared with %s arguments. Please declare with 1 or 2 arguments." % (func.__name__, arguments_count))
    return arguments_count


def __get_arguments_count_of_test(func):
    return len(inspect.getargspec(func)[0])


import os
import traceback
import uuid
from io import BytesIO

from . import config
from .exceptions import ScreenshotError
from .plogger import preporter
from .utils import escape_filename

# ----------------------------------------------------------------------
# -------- [ cross-platform multiple screenshots module ] --------------
# ----------------------------------------------------------------------
'''
    Copyright (c) 2013-2015, Mickael 'Tiger-222' Schoentgen

    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee or royalty is hereby
    granted, provided that the above copyright notice appear in all copies
    and that both that copyright notice and this permission notice appear
    in supporting documentation or portions thereof, including
    modifications, that you make.
'''
from struct import pack
from platform import system
from zlib import compress, crc32

if system() == 'Darwin':
    try:
        import Quartz
        from LaunchServices import kUTTypePNG
        pyobjc_installed = True
    except ImportError:
        pyobjc_installed = False
elif system() == 'Windows':
    from ctypes import c_void_p, create_string_buffer, sizeof, \
        windll, Structure, POINTER, WINFUNCTYPE
    from ctypes.wintypes import BOOL, DOUBLE, DWORD, HBITMAP, HDC, \
        HGDIOBJ, HWND, INT, LPARAM, LONG, RECT, UINT, WORD


    class BITMAPINFOHEADER(Structure):
        _fields_ = [('biSize', DWORD), ('biWidth', LONG), ('biHeight', LONG),
                    ('biPlanes', WORD), ('biBitCount', WORD),
                    ('biCompression', DWORD), ('biSizeImage', DWORD),
                    ('biXPelsPerMeter', LONG), ('biYPelsPerMeter', LONG),
                    ('biClrUsed', DWORD), ('biClrImportant', DWORD)]


    class BITMAPINFO(Structure):
        _fields_ = [('bmiHeader', BITMAPINFOHEADER), ('bmiColors', DWORD * 3)]


class MSS(object):
    def enum_display_monitors(self, screen=0):
        raise NotImplementedError('MSS: subclasses need to implement this!')

    def get_pixels(self, monitor):
        raise NotImplementedError('MSS: subclasses need to implement this!')

    def save(self, output, screen=0):
        for i, monitor in enumerate(self.enum_display_monitors(screen)):
            if screen <= 0 or (screen > 0 and i + 1 == screen):
                self.save_img(data=self.get_pixels(monitor),
                              width=monitor[b'width'],
                              height=monitor[b'height'],
                              output=output)

    def save_img(self, data, width, height, output):
        zcrc32 = crc32
        zcompr = compress
        len_sl = width * 3
        scanlines = b''.join(
            [pack(b'>B', 0) + data[y * len_sl:y * len_sl + len_sl]
             for y in range(height)])

        magic = pack(b'>8B', 137, 80, 78, 71, 13, 10, 26, 10)

        # Header: size, marker, data, CRC32
        ihdr = [b'', b'IHDR', b'', b'']
        ihdr[2] = pack(b'>2I5B', width, height, 8, 2, 0, 0, 0)
        ihdr[3] = pack(b'>I', zcrc32(b''.join(ihdr[1:3])) & 0xffffffff)
        ihdr[0] = pack(b'>I', len(ihdr[2]))

        # Data: size, marker, data, CRC32
        idat = [b'', b'IDAT', b'', b'']
        idat[2] = zcompr(scanlines, 9)
        idat[3] = pack(b'>I', zcrc32(b''.join(idat[1:3])) & 0xffffffff)
        idat[0] = pack(b'>I', len(idat[2]))

        # Footer: size, marker, None, CRC32
        iend = [b'', b'IEND', b'', b'']
        iend[3] = pack(b'>I', zcrc32(iend[1]) & 0xffffffff)
        iend[0] = pack(b'>I', len(iend[2]))

        try:
            output.write(magic + b''.join(ihdr) + b''.join(idat) + b''.join(iend))
        except:
            err = 'MSS: error writing data to "{0}".'.format(output)
            raise ScreenshotError(err)


class MSSMac(MSS):
    def enum_display_monitors(self, screen=0):
        if screen == -1:
            rect = Quartz.CGRectInfinite
            yield ({
                b'left': int(rect.origin.x),
                b'top': int(rect.origin.y),
                b'width': int(rect.size.width),
                b'height': int(rect.size.height)
            })
        else:
            max_displays = 32  # Could be augmented, if needed ...
            rotations = {0.0: 'normal', 90.0: 'right', -90.0: 'left'}
            _, ids, _ = Quartz.CGGetActiveDisplayList(max_displays, None, None)
            for display in ids:
                rect = Quartz.CGRectStandardize(Quartz.CGDisplayBounds(display))
                left, top = rect.origin.x, rect.origin.y
                width, height = rect.size.width, rect.size.height
                rot = Quartz.CGDisplayRotation(display)
                if rotations[rot] in ['left', 'right']:
                    width, height = height, width
                yield ({
                    b'left': int(left),
                    b'top': int(top),
                    b'width': int(width),
                    b'height': int(height)
                })

    def get_pixels(self, monitor):
        width, height = monitor[b'width'], monitor[b'height']
        left, top = monitor[b'left'], monitor[b'top']
        rect = Quartz.CGRect((left, top), (width, height))
        options = Quartz.kCGWindowListOptionOnScreenOnly
        winid = Quartz.kCGNullWindowID
        default = Quartz.kCGWindowImageDefault
        self.image = Quartz.CGWindowListCreateImage(rect, options, winid, default)
        if not self.image:
            raise ScreenshotError('MSS: CGWindowListCreateImage() failed.')
        return self.image

    def save_img(self, data, width, height, output):
        cf_data = Quartz.CFDataCreateMutable(Quartz.kCFAllocatorDefault, 0)
        dest = Quartz.CGImageDestinationCreateWithData(cf_data, kUTTypePNG, 1, None)
        if not dest:
            err = 'MSS: CGImageDestinationCreateWithURL() failed.'
            raise ScreenshotError(err)

        Quartz.CGImageDestinationAddImage(dest, data, None)
        if not Quartz.CGImageDestinationFinalize(dest):
            raise ScreenshotError('MSS: CGImageDestinationFinalize() failed.')

        cf_data_bytes = Quartz.CFDataGetBytes(cf_data, Quartz.CFRangeMake(0, Quartz.CFDataGetLength(cf_data)), None)
        output.write(cf_data_bytes)


class MSSWindows(MSS):
    def __init__(self):
        self._set_argtypes()
        self._set_restypes()

    def _set_argtypes(self):
        self.MONITORENUMPROC = WINFUNCTYPE(INT, DWORD, DWORD, POINTER(RECT),
                                           DOUBLE)
        windll.user32.GetSystemMetrics.argtypes = [INT]
        windll.user32.EnumDisplayMonitors.argtypes = [HDC, c_void_p,
                                                      self.MONITORENUMPROC,
                                                      LPARAM]
        windll.user32.GetWindowDC.argtypes = [HWND]
        windll.gdi32.CreateCompatibleDC.argtypes = [HDC]
        windll.gdi32.CreateCompatibleBitmap.argtypes = [HDC, INT, INT]
        windll.gdi32.SelectObject.argtypes = [HDC, HGDIOBJ]
        windll.gdi32.BitBlt.argtypes = [HDC, INT, INT, INT, INT, HDC, INT, INT,
                                        DWORD]
        windll.gdi32.DeleteObject.argtypes = [HGDIOBJ]
        windll.gdi32.GetDIBits.argtypes = [HDC, HBITMAP, UINT, UINT, c_void_p,
                                           POINTER(BITMAPINFO), UINT]

    def _set_restypes(self):
        windll.user32.GetSystemMetrics.restypes = INT
        windll.user32.EnumDisplayMonitors.restypes = BOOL
        windll.user32.GetWindowDC.restypes = HDC
        windll.gdi32.CreateCompatibleDC.restypes = HDC
        windll.gdi32.CreateCompatibleBitmap.restypes = HBITMAP
        windll.gdi32.SelectObject.restypes = HGDIOBJ
        windll.gdi32.BitBlt.restypes = BOOL
        windll.gdi32.GetDIBits.restypes = INT
        windll.gdi32.DeleteObject.restypes = BOOL

    def enum_display_monitors(self, screen=-1):
        if screen == -1:
            SM_XVIRTUALSCREEN, SM_YVIRTUALSCREEN = 76, 77
            SM_CXVIRTUALSCREEN, SM_CYVIRTUALSCREEN = 78, 79
            left = windll.user32.GetSystemMetrics(SM_XVIRTUALSCREEN)
            right = windll.user32.GetSystemMetrics(SM_CXVIRTUALSCREEN)
            top = windll.user32.GetSystemMetrics(SM_YVIRTUALSCREEN)
            bottom = windll.user32.GetSystemMetrics(SM_CYVIRTUALSCREEN)
            yield ({
                b'left': int(left),
                b'top': int(top),
                b'width': int(right),
                b'height': int(bottom)
            })
        else:

            def _callback(monitor, dc, rect, data):
                rct = rect.contents
                monitors.append({
                    b'left': int(rct.left),
                    b'top': int(rct.top),
                    b'width': int(rct.right - rct.left),
                    b'height': int(rct.bottom - rct.top)
                })
                return 1

            monitors = []
            callback = self.MONITORENUMPROC(_callback)
            windll.user32.EnumDisplayMonitors(0, 0, callback, 0)
            for mon in monitors:
                yield mon

    def get_pixels(self, monitor):
        width, height = monitor[b'width'], monitor[b'height']
        left, top = monitor[b'left'], monitor[b'top']
        SRCCOPY = 0xCC0020
        DIB_RGB_COLORS = BI_RGB = 0
        srcdc = memdc = bmp = None

        try:
            bmi = BITMAPINFO()
            bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER)
            bmi.bmiHeader.biWidth = width
            bmi.bmiHeader.biHeight = -height  # Why minus? See [1]
            bmi.bmiHeader.biPlanes = 1  # Always 1
            bmi.bmiHeader.biBitCount = 24
            bmi.bmiHeader.biCompression = BI_RGB
            buffer_len = height * width * 3
            self.image = create_string_buffer(buffer_len)
            srcdc = windll.user32.GetWindowDC(0)
            memdc = windll.gdi32.CreateCompatibleDC(srcdc)
            bmp = windll.gdi32.CreateCompatibleBitmap(srcdc, width, height)
            windll.gdi32.SelectObject(memdc, bmp)
            windll.gdi32.BitBlt(memdc, 0, 0, width, height, srcdc, left, top,
                                SRCCOPY)
            bits = windll.gdi32.GetDIBits(memdc, bmp, 0, height, self.image,
                                          bmi, DIB_RGB_COLORS)
            if bits != height:
                raise ScreenshotError('MSS: GetDIBits() failed.')
        finally:
            # Clean up
            if srcdc:
                windll.gdi32.DeleteObject(srcdc)
            if memdc:
                windll.gdi32.DeleteObject(memdc)
            if bmp:
                windll.gdi32.DeleteObject(bmp)

        # Replace pixels values: BGR to RGB
        self.image[2:buffer_len:3], self.image[0:buffer_len:3] = \
            self.image[0:buffer_len:3], self.image[2:buffer_len:3]
        return self.image


def mss(*args, **kwargs):
    mss_class = {
        'Darwin': MSSMac,
        'Windows': MSSWindows
    }[system()]

    return mss_class(*args, **kwargs)


# ----------------------------------------------------------------------
# ----------- [ take screenshot for webdriver or desktop ] -------------
# ----------------------------------------------------------------------
def take_screenshots():
    from . import testexecutor

    web_drivers = testexecutor.current_executor().get_property("web_drivers")
    running_test_fixture = testexecutor.current_executor().get_property("running_test_fixture")
    hash_code = str(uuid.uuid4()).split("-")[0]
    screenshots = []

    screenshot = {
        "source": "Desktop",
        "path": "%s-%s-0.png" % (escape_filename(running_test_fixture.full_name), hash_code)
    }

    if system() == 'Darwin' and not pyobjc_installed:
        screenshot["error"] = "The package pyobjc is necessary for taking screenshot of desktop, please install it."
    else:
        try:
            output = BytesIO()
            mss().save(output=output, screen=-1)  # -1 means all monitors
            value = output.getvalue()
            with open(os.path.join(config.get_option("temp"), screenshot["path"]), mode="wb") as f:
                f.write(value)
        except Exception as e:
            screenshot["error"] = str(e).strip() or "\n".join([str(arg) for arg in e.args])

    screenshots.append(screenshot)

    if web_drivers:
        for index, web_driver in enumerate(web_drivers):
            screenshot = {
                "source": "Web Driver",
                "path": "%s-%s-%s.png" % (escape_filename(running_test_fixture.full_name), hash_code, index + 1)
            }

            try:
                screenshot["alert"] = web_driver.switch_to.alert.text
            except Exception as e:
                pass

            while True:
                try:
                    web_driver.switch_to.alert.dismiss()
                except Exception as e:
                    break

            try:
                screenshot["url"] = web_driver.current_url
            except Exception as e:
                pass

            try:
                screenshot["title"] = web_driver.title
            except Exception as e:
                pass

            try:
                value = web_driver.get_screenshot_as_png()
                with open(os.path.join(config.get_option("temp"), screenshot["path"]), mode="wb") as f:
                    f.write(value)
            except Exception as e:
                screenshot["error"] = str(e).strip() or "\n".join([str(arg) for arg in e.args])

            screenshots.append(screenshot)

    return screenshots


from ptest.decorator import TestClass, Test, BeforeMethod, AfterMethod
from ptest.assertion import assert_equals
from ptest.plogger import preporter


class TestBase:
    @BeforeMethod()
    def before(self):
        preporter.info("setting expected result.")
        self.expected = 10

    @AfterMethod(always_run=True)
    def after(self):
        preporter.info("cleaning up")


@TestClass()
class PTestClass(TestBase):
    @Test()
    def test1(self):
        assert_equals(10, self.expected)  # pass

    @Test()
    def test2(self):
        assert_equals(20, self.expected)  # failed



from ptest.decorator import TestClass, Test, BeforeMethod, AfterMethod, BeforeSuite, BeforeClass, BeforeGroup, \
    AfterGroup, AfterClass, AfterSuite
from ptest.assertion import assert_equals, fail, assert_not_none
from ptest.plogger import preporter
from ptest import config


# the test cases in this class will be executed by multiple threads
@TestClass(run_mode="parallel")
class PTestClass:
    @BeforeSuite()
    def before_suite(self):
        preporter.info("before suite")

    @BeforeClass()
    def before_class(self):
        preporter.info("before class")

    @BeforeGroup()
    def before_group(self):
        preporter.info("before group")

    @BeforeMethod(description="Prepare test data.")
    def before(self):
        preporter.info("setting expected result.")
        self.expected = 10

    @Test(tags=["regression", "smoke"])
    def test1(self):
        assert_equals(10, self.expected)  # pass

    @Test(tags="smoke, nightly")
    def test2(self):
        # assert the property defined via -D<key>=<value> in cmd line
        assert_not_none(config.get_property("key"))

    @Test(enabled=False)  # won't be run
    def test3(self):
        fail("failed")

    # always_run means that the @AfterMethod will be run even the @BeforeMethod failed
    @AfterMethod(always_run=True, description="Clean up")
    def after(self):
        preporter.info("cleaning up")

    @AfterGroup()
    def after_group(self):
        preporter.info("after group")

    @AfterClass()
    def after_class(self):
        preporter.info("after class")

    @AfterSuite()
    def after_suite(self):
        preporter.info("after suite")



import re

from .utils import StringTypes


# -------------------------------------------
# ------------ simple assertion -------------
# -------------------------------------------
def assert_true(actual, msg=""):
    if actual is not True:
        __raise_error(msg, "Expected: <True>, Actual: <%s>." % actual)


def assert_false(actual, msg=""):
    if actual is not False:
        __raise_error(msg, "Expected: <False>, Actual: <%s>." % actual)


def assert_none(actual, msg=""):
    if actual is not None:
        __raise_error(msg, "Expected: <None>, Actual: <%s>." % actual)


def assert_not_none(actual, msg=""):
    if actual is None:
        __raise_error(msg, "Expected: NOT <None>, Actual: <None>.")


def assert_equals(actual, expected, msg=""):
    if not actual == expected:
        __raise_error(msg, "Expected: <%s>, Actual: <%s>." % (expected, actual))


def assert_not_equals(actual, not_expected, msg=""):
    if actual == not_expected:
        __raise_error(msg, "Expected: NOT <%s>, Actual: <%s>." % (not_expected, actual))


def assert_list_equals(actual_list, expected_list, msg=""):
    if len(actual_list) != len(expected_list):
        __raise_error(msg, "size of expected list <%s> is: <%s>, but size of actual list <%s> is: <%s>." %
                      (expected_list, len(expected_list), actual_list, len(actual_list)))
    for i, element in enumerate(actual_list):
        if not element == expected_list[i]:
            __raise_error(msg, "element <index: %s> of expected list <%s> is: <%s>, but element <index: %s> of actual list <%s> is: <%s>" %
                    (i, expected_list, expected_list[i], i, actual_list, actual_list[i]))


def assert_list_elements_equal(actual_list, expected_list, msg=""):
    diff_elements = [element for element in actual_list if not element in expected_list]
    if len(diff_elements) != 0:
        __raise_error(msg, "expected list <%s> doesn't contain elements <%s> from actual list <%s>." % (expected_list, diff_elements, actual_list))
    diff_elements = [element for element in expected_list if not element in actual_list]
    if len(diff_elements) != 0:
        __raise_error(msg, "actual list <%s> doesn't contain elements <%s> from expected list <%s>." % (actual_list, diff_elements, expected_list))


def assert_set_contains(superset, subset, msg=""):
    diff_elements = [element for element in subset if not element in superset]
    if len(diff_elements) != 0:
        __raise_error(msg, "Superset <%s> doesn't contain elements <%s> from subset <%s>." % (superset, diff_elements, subset))


def fail(msg=""):
    __raise_error("", msg)


def __raise_error(msg, error_msg):
    if msg:
        raise_msg = "%s\n%s" % (msg, error_msg)
    else:
        raise_msg = "%s" % error_msg
    raise AssertionError(raise_msg)

# -------------------------------------------
# --------- "assert that" assertion ---------
# -------------------------------------------
from numbers import Number
from datetime import datetime, date


SUBJECT_TYPE_MAP = {}

def assert_that(subject):
    if subject is None:
        return _NoneSubject(subject)
    if isinstance(subject, StringTypes):
        return _StringSubject(subject)
    if isinstance(subject, bool):
        return _BoolSubject(subject)
    if isinstance(subject, Number):
        return _NumericSubject(subject)
    if isinstance(subject, (list, tuple)):
        return _ListOrTupleSubject(subject)
    if isinstance(subject, set):
        return _SetSubject(subject)
    if isinstance(subject, dict):
        return _DictSubject(subject)
    if isinstance(subject, datetime):
        return _DateTimeSubject(subject)
    if isinstance(subject, date):
        return _DateSubject(subject)
    for subject_type, subject_class in SUBJECT_TYPE_MAP.items():
        if isinstance(subject, subject_type):
            return subject_class(subject)
    return _ObjSubject(subject)

# this method is used to format the obj
def _format(obj):
    if isinstance(obj, StringTypes):
        return "\"%s\"" % obj
    if isinstance(obj, list):
        return ", ".join([_format(i) for i in obj])
    return str(obj)

# this method is used to get the type name of the object
def _type(obj):
    return type(obj).__name__

class _Subject(object):
    def __init__(self, subject):
        self._subject = subject
        self._subject_name = None
        self._msg = None

    def named(self, name):
        """
            Give a name to the subject.
        """
        self._subject_name = str(name)
        return self

    def with_message(self, message):
        """
           Set custom message for this assertion.
        """
        self._msg = str(message)
        return self

    def _raise_error(self, partial_error_msg, error=AssertionError):
        if self._subject_name is None:
            error_msg = "Unexpectedly that the %s <%s> %s" % (_type(self._subject), self._subject, partial_error_msg)
        else:
            error_msg = "Unexpectedly that the %s named \"%s\" %s" % (_type(self._subject), self._subject_name, partial_error_msg)
        self._raise_raw_error(error_msg, error)

    def _raise_raw_error(self, error_msg, error=AssertionError):
        if self._msg:
            raise_msg = "%s\n%s" % (self._msg, error_msg)
        else:
            raise_msg = "%s" % error_msg
        raise error(raise_msg)


class _ObjSubject(_Subject):
    def __init__(self, subject):
        _Subject.__init__(self, subject)

    def is_instance_of(self, class_or_type_or_tuple):
        """
            Fails if the subject is not an instance of given class or type or tuple of types.

            Assert whether an object is an instance of a class or of a subclass thereof.
            With a type as second argument, return whether that is the object's type.
            The form using a tuple, is_instance_of(x, (A, B, ...)), is a shortcut for
            is_instance_of(x, A) or is_instance_of(x, B) or ... (etc.).
        """
        if not isinstance(self._subject, class_or_type_or_tuple):
            self._raise_error("is not instance of <%s>." % _format(class_or_type_or_tuple))
        return self

    def is_type_of(self, type_):
        """
           Fails if the subject is not of given type.
        """
        if type(self._subject) is not type_:
            self._raise_error("is not type of %s." % type_)
        return self

    def is_equal_to(self, other_obj):
        """
            Fails if the subject is not equal to other obj.
        """
        if not self._subject == other_obj:
            self._raise_error("is not equal to %s <%s>." % (_type(other_obj), other_obj))
        return self

    def is_not_equal_to(self, other_obj):
        """
            Fails if the subject is equal to other obj.
        """
        if self._subject == other_obj:
            self._raise_error("is equal to %s <%s>." % (_type(other_obj), other_obj))
        return self

    def is_same_as(self, other_obj):
        """
            Fails if the subject is not identical to other obj.
        """
        if self._subject is not other_obj:
            self._raise_error("is not identical to %s <%s>." % (_type(other_obj), other_obj))
        return self

    def is_not_same_as(self, other_obj):
        """
            Fails if the subject is identical to other obj.
        """
        if self._subject is other_obj:
            self._raise_error("is identical to %s <%s>." % (_type(other_obj), other_obj))
        return self

    def is_none(self):
        """
            Fails if the subject is not None.
        """
        if self._subject is not None:
            self._raise_error("is not <None>.")
        return self

    def is_not_none(self):
        """
            Fails if the subject is None.
        """
        if self._subject is None:
            self._raise_error("is <None>.")
        return self

    def is_in(self, iterable):
        """
           Fails unless the subject is equal to any element in the given iterable.
        """
        if self._subject not in iterable:
            self._raise_error("is not in %s <%s>." % (_type(iterable), iterable))
        return self

    def is_not_in(self, iterable):
        """
           Fails if the subject is equal to any element in the given iterable.
        """
        if self._subject in iterable:
            self._raise_error("is in %s <%s>." % (_type(iterable), iterable))
        return self

    def meets(self, func):
        """
            Fails if the subject doesn't meet the given function.
            Note: The function must accepts one argument.

            For Example:
                assert_that(99).meets(lambda x: x > 0)

                def is_positive(num):
                    return num > 0
                assert_that(99).meets(is_positive)
        """
        if not func(self._subject):
            self._raise_error("doesn't meet function <%s>." % func.__name__)
        return self

    def attr(self, attribute_name):
        """
            Assert the attribute of this subject. If the attribute does not exist, raise AttributeError.
        """
        if not hasattr(self._subject, attribute_name):
            self._raise_error("doesn't have attribute <%s>." % attribute_name, error=AttributeError)
        return assert_that(getattr(self._subject, attribute_name))

    def has_attr(self, attribute_name):
        """
            Fails if the subject doesn't have the given attribute.
        """
        if not hasattr(self._subject, attribute_name):
            self._raise_error("doesn't have attribute <%s>." % attribute_name)
        return self

    def does_not_have_attr(self, attribute_name):
        """
            Fails if the subject has the given attribute.
        """
        if hasattr(self._subject, attribute_name):
            self._raise_error("has attribute <%s>." % attribute_name)
        return self

    def __getattr__(self, item):
        if self._subject_name is None:
            self._raise_raw_error("Cannot perform assertion \"%s\" for %s <%s>." % (item, _type(self._subject), self._subject), error=AttributeError)
        else:
            self._raise_raw_error("Cannot perform assertion \"%s\" for %s named \"%s\"." % (item, _type(self._subject), self._subject_name), error=AttributeError)


class _NoneSubject(_ObjSubject):
    def __init__(self, subject):
        _ObjSubject.__init__(self, subject)


class _BoolSubject(_ObjSubject):
    def __init__(self, subject):
        _ObjSubject.__init__(self, subject)

    def is_true(self):
        """
            Fails if the subject is false.
        """
        if self._subject is not True:
            self._raise_error("is not <True>.")
        return self

    def is_false(self):
        """
            Fails if the subject is true.
        """
        if self._subject is not False:
            self._raise_error("is not <False>.")
        return self


class _NumericSubject(_ObjSubject):
    def __init__(self, subject):
        _ObjSubject.__init__(self, subject)

    def is_less_than(self, other_number):
        """
            Fails if the subject is not less than other number.
        """
        if self._subject >= other_number:
            self._raise_error("is not less than %s <%s>." % (_type(other_number), other_number))
        return self

    def is_greater_than(self, other_number):
        """
            Fails if the subject is not greater than other number.
        """
        if self._subject <= other_number:
            self._raise_error("is not greater than %s <%s>." % (_type(other_number), other_number))
        return self

    def is_less_than_or_equal_to(self, other_number):
        """
            Fails if the subject is greater than other number.
        """
        if self._subject > other_number:
            self._raise_error("is greater than %s <%s>." % (_type(other_number), other_number))
        return self

    def is_at_most(self, other_number):
        """
            Fails if the subject is greater than other number.
        """
        return self.is_less_than_or_equal_to(other_number)

    def is_greater_than_or_equal_to(self, other_number):
        """
            Fails if the subject is less than other number.
        """
        if self._subject < other_number:
            self._raise_error("is less than %s <%s>." % (_type(other_number), other_number))
        return self

    def is_at_least(self, other_number):
        """
            Fails if the subject is less than other number.
        """
        return self.is_greater_than_or_equal_to(other_number)

    def is_zero(self):
        """
            Fails if the subject is not zero (0).
        """
        if self._subject != 0:
            self._raise_error("is not <0>.")
        return self

    def is_not_zero(self):
        """
            Fails if the subject is zero (0).
        """
        if self._subject == 0:
            self._raise_error("is <0>.")
        return self

    def is_positive(self):
        """
            Fails if the subject is not positive.
        """
        if self._subject <= 0:
            self._raise_error("is not positive.")
        return self

    def is_negative(self):
        """
            Fails if the subject is not negative.
        """
        if self._subject >= 0:
            self._raise_error("is not negative.")
        return self

    def is_between(self, low, high):
        """
            Fails if the subject is not between low and high.

            Note: low and high are included
        """
        if self._subject < low or self._subject > high:
            self._raise_error("is not between low %s <%s> and high %s <%s>." % (_type(low), low, _type(high), high))
        return self


class _IterableSubject(_ObjSubject):
    def __init__(self, subject):
        _ObjSubject.__init__(self, subject)

    def is_empty(self):
        """
            Fails if the subject is empty.
        """
        if len(self._subject) != 0:
            self._raise_error("is not empty.")
        return self

    def is_not_empty(self):
        """
            Fails if the subject is not empty.
        """
        if len(self._subject) == 0:
            self._raise_error("is empty.")
        return self

    def has_length(self, expected_length):
        """
            Fails if the subject does not have the given length.
        """
        if not len(self._subject) == expected_length:
            self._raise_error("doesn't have a length of <%s>. It is <%s>." % (expected_length, len(self._subject)))
        return self

    def contains(self, obj):
        """
            Fails if the subject doesn't contain the given object.
        """
        if obj not in self._subject:
            self._raise_error("doesn't contain %s <%s>." % (_type(obj), obj))
        return self

    def does_not_contain(self, obj):
        """
            Fails if the subject contain the given object.
        """
        if obj in self._subject:
            self._raise_error("contains %s <%s>." % (_type(obj), obj))
        return self

    def contains_all_of(self, *objs):
        """
            Fails unless the subject contains all of the given objects.
        """
        uncontained_objs = [obj for obj in objs if obj not in self._subject]
        if uncontained_objs:
            self._raise_error("doesn't contain elements <%s> in <%s>." % (_format(uncontained_objs), _format(list(objs))))
        return self

    def contains_all_in(self, iterable):
        """
            Fails unless the subject contains all in the given iterable.
        """
        uncontained_objs = [obj for obj in iterable if obj not in self._subject]
        if uncontained_objs:
            self._raise_error("doesn't contain elements <%s> in %s <%s>." % (_format(uncontained_objs), _type(iterable), iterable))
        return self

    def contains_any_of(self, *objs):
        """
            Fails unless the subject contains any of the given objects.
        """
        contained_objs = [obj for obj in objs if obj in self._subject]
        if not contained_objs:
            self._raise_error("doesn't contain any element in <%s>." % _format(list(objs)))
        return self

    def contains_any_in(self, iterable):
        """
            Fails unless the subject contains any in the given iterable.
        """
        contained_objs = [obj for obj in iterable if obj in self._subject]
        if not contained_objs:
            self._raise_error("doesn't contain any element in %s <%s>." % (_type(iterable), iterable))
        return self

    def contains_none_of(self, *objs):
        """
            Fails if the subject contains any of the given objects.
        """
        contained_objs = [obj for obj in objs if obj in self._subject]
        if contained_objs:
            self._raise_error("contains elements <%s> in <%s>." % (_format(contained_objs), _format(list(objs))))
        return self

    def contains_none_in(self, iterable):
        """
            Fails if the subject contains any in the given iterable.
        """
        contained_objs = [obj for obj in iterable if obj in self._subject]
        if contained_objs:
            self._raise_error("contains elements <%s> in %s <%s>." % (_format(contained_objs), _type(iterable), iterable))
        return self

    def length(self):
        """
            Assert the length of this subject.
        """
        return _NumericSubject(len(self._subject))

    def each(self):
        """
            For each obj in this subject.
        """
        return _IterableEachSubject(self._subject)


class _IterableEachSubject(object):
    def __init__(self, iterable_subject):
        self._iterable_subject = iterable_subject

    def __getattr__(self, item):
        if item in ["length", "index", "key", "attr"]:
            def each_attr(*args, **kwargs):
                iterable_subject = []
                for subject in self._iterable_subject:
                    iterable_subject.append(getattr(assert_that(subject), item)(*args, **kwargs)._subject)
                return _IterableEachSubject(iterable_subject)

            return each_attr
        elif item in ["each", "each_key", "each_value"]:
            def each_each(*args, **kwargs):
                iterable_subject = []
                for iterable in self._iterable_subject:
                    for subject in getattr(assert_that(iterable), item)(*args, **kwargs)._iterable_subject:
                        iterable_subject.append(subject)
                return _IterableEachSubject(iterable_subject)

            return each_each
        else:
            def assert_each(*args, **kwargs):
                for subject in self._iterable_subject:
                    getattr(assert_that(subject), item)(*args, **kwargs)
                return self

            return assert_each


class _StringSubject(_IterableSubject):
    def __init__(self, subject):
        _IterableSubject.__init__(self, subject)

    def is_equal_to_ignoring_case(self, string):
        """
            Fails if the string is not equal to other string ignoring case.
        """
        if not self._subject.lower() == string.lower():
            self._raise_error("is not equal to %s <%s> ignoring case." % (_type(string), string))
        return self

    def is_blank(self):
        """
            Fails if the string is not blank.
        """
        if len(self._subject.strip()) != 0:
            self._raise_error("is not blank.")
        return self

    def is_not_blank(self):
        """
            Fails if the string is blank.
        """
        if len(self._subject.strip()) == 0:
            self._raise_error("is blank.")
        return self

    def starts_with(self, prefix):
        """
            Fails if the string does not start with the given string.
        """
        if not self._subject.startswith(prefix):
            self._raise_error("doesn't start with %s <%s>." % (_type(prefix), prefix))
        return self

    def ends_with(self, suffix):
        """
            Fails if the string does not end with the given string.
        """
        if not self._subject.endswith(suffix):
            self._raise_error("doesn't end with %s <%s>." % (_type(suffix), suffix))
        return self

    def matches(self, regex):
        """
            Fails if the string doesn't match the given regex.

            Note: If you want to match the entire string, just include anchors in the regex pattern.
        """
        if not re.compile(regex).search(self._subject):
            self._raise_error("doesn't match regex <%s>." % regex)
        return self

    def does_not_match(self, regex):
        """
            Fails if the string match the given regex.

            Note: If you want to match the entire string, just include anchors in the regex pattern.
        """
        if re.compile(regex).search(self._subject):
            self._raise_error("matches regex <%s>." % regex)
        return self

    def is_alpha(self):
        """
            Fails if the string doesn't contain only alphabetic chars.
        """
        if not self._subject.isalpha():
            self._raise_error("doesn't contain only alphabetic chars.")
        return self

    def is_digit(self):
        """
            Fails if the string doesn't contain only digits.
        """
        if not self._subject.isdigit():
            self._raise_error("doesn't contain only digits.")
        return self

    def is_lower(self):
        """
            Fails if the string doesn't contain only lowercase chars.
        """
        if not self._subject == self._subject.lower():
            self._raise_error("doesn't contain only lowercase chars.")
        return self

    def is_upper(self):
        """
            Fails if the string doesn't contain only uppercase chars.
        """
        if not self._subject == self._subject.upper():
            self._raise_error("doesn't contain only uppercase chars.")
        return self


class _ListOrTupleSubject(_IterableSubject):
    def __init__(self, subject):
        _IterableSubject.__init__(self, subject)

    def has_same_elements_as(self, other_list_or_tuple):
        """
            Fails unless this list/tuple has the same elements as other list/tuple.
        """
        uncontained_objs = [obj for obj in other_list_or_tuple if obj not in self._subject]
        if uncontained_objs:
            self._raise_error("doesn't contain elements <%s> in %s <%s>." %
                              (_format(uncontained_objs), _type(other_list_or_tuple), other_list_or_tuple))
            
        uncontained_objs = [obj for obj in self._subject if obj not in other_list_or_tuple]
        if uncontained_objs:
            self._raise_error("contains elements <%s> not in %s <%s>." %
                              (_format(uncontained_objs), _type(other_list_or_tuple), other_list_or_tuple))
        return self

    def contains_duplicates(self):
        """
            Fails if this list/tuple does not contain duplicate elements.
        """
        if len(self._subject) == len(set(self._subject)):
            self._raise_error("doesn't contain duplicate elements.")
        return self

    def does_not_contain_duplicates(self):
        """
            Fails if this list/tuple contains duplicate elements.
        """
        element_counter = {}
        for element in self._subject:
            if element in element_counter:
                element_counter[element] += 1
            else:
                element_counter[element] = 1
        duplicates = [element for element, count in element_counter.items() if count > 1]
        if duplicates:
            self._raise_error("contains duplicate elements <%s>." % _format(duplicates))
        return self

    def index(self, index):
        """
            Assert the obj of this list/tuple by index. If index doesn't exist, raise IndexError.
        """
        if index >= len(self._subject) or index < 0:
            self._raise_error("has no object of index <%s>." % index, error=IndexError)
        return assert_that(self._subject[index])


class _SetSubject(_IterableSubject):
    def __init__(self, subject):
        _IterableSubject.__init__(self, subject)
    
    def is_super_of(self, other_set):
        """
            Fails unless this set is a super set of other set.
        """
        uncontained_objs = [obj for obj in other_set if obj not in self._subject]
        if uncontained_objs:
            self._raise_error("doesn't contain elements <%s> in %s <%s>." %
                              (_format(uncontained_objs), _type(other_set), other_set))
        return self

    def is_sub_of(self, other_set):
        """
            Fails unless this set is a sub set of other set.
        """
        uncontained_objs = [obj for obj in self._subject if obj not in other_set]
        if uncontained_objs:
            self._raise_error("contains elements <%s> not in %s <%s>." %
                              (_format(uncontained_objs), _type(other_set), other_set))
        return self


class _DictSubject(_IterableSubject):
    def __init__(self, subject):
        _IterableSubject.__init__(self, subject)

    def contains_key(self, key):
        """
            Fails if the dict doesn't contain the given key.
        """
        if key not in self._subject:
            self._raise_error("doesn't contain key %s <%s>." % (_type(key), key))
        return self

    def does_not_contain_key(self, key):
        """
            Fails if the dict contains the given key.
        """
        if key in self._subject:
            self._raise_error("contains key %s <%s>." % (_type(key), key))
        return self

    def contains_value(self, value):
        """
            Fails if the dict doesn't contain the given value.
        """
        if value not in self._subject.values():
            self._raise_error("doesn't contain value %s <%s>." % (_type(value), value))
        return self

    def does_not_contain_value(self, value):
        """
            Fails if the dict contains the given value.
        """
        if value in self._subject.values():
            self._raise_error("contains value %s <%s>." % (_type(value), value))
        return self

    def contains_entry(self, key, value):
        """
           Fails if the dict doesn't contain the given entry.
        """
        if (key, value) not in self._subject.items():
            self._raise_error("doesn't contain entry, key: %s <%s>, value: %s <%s>." % (_type(key), key, _type(value), value))
        return self

    def does_not_contain_entry(self, key, value):
        """
            Fails if the dict contain the given entry.
        """
        if (key, value) in self._subject.items():
            self._raise_error("contains entry, key: %s <%s>, value: %s <%s>." % (_type(key), key, _type(value), value))
        return self

    def is_super_of(self, other_dict):
        """
            Fails unless this dict contains all entries in other dict.
        """
        uncontained_entries = [entry for entry in other_dict.items() if entry not in self._subject.items()]
        if uncontained_entries:
            self._raise_error("doesn't contain entries <%s> in %s <%s>." %
                              (_format(uncontained_entries), _type(other_dict), other_dict))
        return self

    def is_sub_of(self, other_dict):
        """
            Fails unless all entries of this dict are in other dict.
        """
        uncontained_entries = [entry for entry in self._subject.items() if entry not in other_dict.items()]
        if uncontained_entries:
            self._raise_error("contains entries <%s> not in %s <%s>." %
                              (_format(uncontained_entries), _type(other_dict), other_dict))
        return self

    def key(self, key):
        """
            Assert the value of this dict by key. If key doesn't exist, raise KeyError
        """
        if key not in self._subject:
            self._raise_error("doesn't contain key %s <%s>." % (_type(key), key), error=KeyError)
        return assert_that(self._subject[key])

    def each(self):
        """
            For each entry in this dict.
        """
        return _IterableEachSubject(self._subject.items())

    def each_key(self):
        """
            For each key in this dict.
        """
        return _IterableEachSubject(self._subject.keys())

    def each_value(self):
        """
            For each value in this dict.
        """
        return _IterableEachSubject(self._subject.values())


class _DateSubject(_ObjSubject):
    def __init__(self, subject):
        _ObjSubject.__init__(self, subject)

    def is_before(self, other_date):
        """
            Fails if this date is not before other date.
        """
        if self._subject >= other_date:
            self._raise_error("is not before %s <%s>." % (_type(other_date), other_date))
        return self

    def is_after(self, other_date):
        """
            Fails if the date is not after other date.
        """
        if self._subject <= other_date:
            self._raise_error("is not after %s <%s>." % (_type(other_date), other_date))
        return self


class _DateTimeSubject(_DateSubject):
    def __init__(self, subject):
        _DateSubject.__init__(self, subject)




import os
import platform
import re
from codecs import open
from optparse import OptionParser, OptionGroup

from . import __version__

_properties = {}
_options = {}


def get_option(option):
    try:
        return _options[option]
    except KeyError:
        return None


def get_property(key, default=None):
    """
        Get property value.
        If no property found, default value will be returned.
    """
    try:
        return _properties[key]
    except KeyError:
        return default


def get_int_property(key, default=None):
    """
        Get property value and convert it to int.
        If no property found, default value will be returned.
    """
    try:
        return int(_properties[key])
    except KeyError:
        return default


def get_float_property(key, default=None):
    """
        Get property value and convert it to float.
        If no property found, default value will be returned.
    """
    try:
        return float(_properties[key])
    except KeyError:
        return default


def get_boolean_property(key, default=None):
    """
        Get property value and convert it to boolean.
        If no property found, default value will be returned.
    """
    try:
        value = _properties[key]
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        raise ValueError("could not convert string to boolean: %s" % value)
    except KeyError:
        return default


def get_list_property(key, default=None, sep=","):
    """
        Get property value and convert it to list.
        If no property found, default value will be returned.
    """
    try:
        return _properties[key].split(sep)
    except KeyError:
        return default


def load(args):
    option_args, property_args = __load_args(args)
    _parse_options(option_args)
    _load_properties_from_file()
    _parse_properties(property_args)


def _load_properties_from_file():
    property_file = get_option("property_file")
    if property_file is not None:
        file_object = open(property_file, encoding="utf-8")
        try:
            property_regex_str = r"^([^;#].*?)=(.*?)$"
            property_regex = re.compile(property_regex_str)
            for line in file_object:
                property_match = property_regex.search(line.strip())
                if property_match:
                    _properties[property_match.group(1)] = property_match.group(2)
        finally:
            file_object.close()


def __load_args(args):
    property_args = []
    option_args = []
    property_regex_str = r"^-D(.*?)=(.*?)$"  # the format of property definition must be -D<key>=<value>
    property_regex = re.compile(property_regex_str)
    for arg in args:
        property_match = property_regex.search(arg)
        if property_match:
            property_args.append(arg)
        else:
            option_args.append(arg)
    return option_args, property_args


def _parse_properties(property_args):
    property_regex_str = r"^-D(.*?)=(.*?)$"  # the format of property definition must be -D<key>=<value>
    property_regex = re.compile(property_regex_str)
    for arg in property_args:
        property_match = property_regex.search(arg)
        _properties[property_match.group(1)] = property_match.group(2)


def _parse_options(option_args):
    parser = OptionParser(usage="ptest [options] [properties]", version="ptest %s for Python %s" % (__version__, platform.python_version()),
                          description="ptest is a light test runner for Python.")

    # path and property
    parser.add_option("-w", "--workspace", action="store", dest="workspace", default=".", metavar="dir",
                      help="Specify the workspace dir (relative to working directory). Default is current working directory.")
    parser.add_option("-P", "--pythonpaths", action="store", dest="python_paths", default=None, metavar="paths",
                      help="Specify the additional locations (relative to workspace) where to search test libraries from when they are imported. "
                           "Multiple paths can be given by separating them with a comma.")
    parser.add_option("-p", "--propertyfile", action="store", dest="property_file", default=None, metavar="file",
                      help="Specify the property file (relative to workspace). "
                           "The properties in property file will be overwritten by user defined properties in cmd line. "
                           "Get property via get_property() in module ptest.config.")

    # running
    parser.add_option("-R", "--runfailed", action="store", dest="run_failed", default=None, metavar="file",
                      help="Specify the xunit result xml path (relative to workspace) and run the failed test cases in it.")
    parser.add_option("-t", "--targets", action="store", dest="test_targets", default=None, metavar="targets",
                      help="Specify the path of test targets, separated by comma. Test target can be package/module/class/method. "
                           "The target path format is: package[.module[.class[.method]]] "
                           "NOTE: ptest ONLY searches modules under --workspace, --pythonpaths and sys.path")
    parser.add_option("-f", "--filter", action="store", dest="test_filter", default=None, metavar="class",
                      help="Specify the path of test filter class, select test cases to run by the specified filter. "
                           "The test filter class should implement class TestFilter in ptest.testfilter "
                           "The filter path format is: package.module.class "
                           "NOTE: ptest ONLY searches modules under --workspace, --pythonpaths and sys.path")
    parser.add_option("-i", "--includetags", action="store", dest="include_tags", default=None, metavar="tags",
                      help="Select test cases to run by tags, separated by comma.")
    parser.add_option("-e", "--excludetags", action="store", dest="exclude_tags", default=None, metavar="tags",
                      help="Select test cases not to run by tags, separated by comma. These test cases are not run even if included with --includetags.")
    parser.add_option("-g", "--includegroups", action="store", dest="include_groups", default=None, metavar="groups",
                      help="Select test cases to run by groups, separated by comma.")
    parser.add_option("-n", "--testexecutornumber", action="store", dest="test_executor_number", metavar="int",
                      default=1, help="Specify the number of test executors. Default value is 1.")

    # output
    parser.add_option("-o", "--outputdir", action="store", dest="output_dir", default="test-output", metavar="dir",
                      help="Specify the output dir (relative to workspace).")
    parser.add_option("-r", "--reportdir", action="store", dest="report_dir", default="html-report", metavar="dir",
                      help="Specify the html report dir (relative to output dir).")
    parser.add_option("-x", "--xunitxml", action="store", dest="xunit_xml", default="xunit-results.xml",
                      metavar="file", help="Specify the xunit result xml path (relative to output dir).")

    # miscellaneous
    parser.add_option("-l", "--listeners", action="store", dest="test_listeners", default=None, metavar="class",
                      help="Specify the path of test listener classes, separated by comma. "
                           "The listener class should implement class TestListener in ptest.plistener "
                           "The listener path format is: package.module.class "
                           "NOTE: 1. ptest ONLY searches modules under --workspace, --pythonpaths and sys.path "
                           "2. The listener class must be thread safe if you set -n(--testexecutornumber) greater than 1.")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False,
                      help="Set ptest console to verbose mode.")
    parser.add_option("--temp", action="store", dest="temp", default="ptest-temp", metavar="dir",
                      help="Specify the temp dir (relative to workspace).")
    parser.add_option("--disablescreenshot", action="store_true", dest="disable_screenshot", default=False,
                      help="Disable taking screenshot for preporter.")

    # tool
    parser.add_option("-m", "--mergexunitxmls", action="store", dest="merge_xunit_xmls", default=None, metavar="files",
                      help="Merge the xunit result xmls (relative to workspace). Multiple files can be given by separating them with a comma."
                            "Use --to to specify the path of merged xunit result xml.")
    parser.add_option("--to", action="store", dest="to", default=None, metavar='path',
                      help="Specify the 'to' destination (relative to workspace).")

    # user defined properties
    parser.add_option_group(
        OptionGroup(parser, "User defined properties",
                    "Define properties via -D<key>=<value>. Get defined property via get_property() in module ptest.config."))

    options, unknown_args = parser.parse_args(option_args)

    # only one of the main options can be specified
    main_options = [options.test_targets, options.run_failed, options.merge_xunit_xmls]
    specified_options_count = len([option for option in main_options if option is not None])
    if specified_options_count == 0:
        parser.error("You must specify one of the following options: -t(--targets), -R(--runfailed), -m(--mergexunitxmls).")
    elif specified_options_count > 1:
        parser.error("You can ONLY specify one of the following options: -t(--targets), -R(--runfailed), -m(--mergexunitxmls).")

    # check '--to'
    if options.merge_xunit_xmls is not None and options.to is None:
        parser.error("You must use --to to specify the path of merged xunit result xml (--mergexunitxmls).")

    # spilt multiple values by comma
    def split(option_value):
        return None if option_value is None else option_value.split(",")

    options.python_paths = split(options.python_paths)
    options.test_targets = split(options.test_targets)
    options.include_tags = split(options.include_tags)
    options.exclude_tags = split(options.exclude_tags)
    options.include_groups = split(options.include_groups)
    options.test_listeners = split(options.test_listeners)
    options.merge_xunit_xmls = split(options.merge_xunit_xmls)

    # convert to full path for options
    def join_path(base_path, sub_path):
        return os.path.abspath(os.path.join(base_path, sub_path))

    options.workspace = join_path(os.getcwd(), options.workspace)
    options.python_paths = None if options.python_paths is None else [join_path(options.workspace, path) for path in options.python_paths]
    options.property_file = None if options.property_file is None else join_path(options.workspace, options.property_file)

    options.run_failed = None if options.run_failed is None else join_path(options.workspace, options.run_failed)
    options.output_dir = join_path(options.workspace, options.output_dir)
    options.report_dir = join_path(options.output_dir, options.report_dir)
    options.xunit_xml = join_path(options.output_dir, options.xunit_xml)
    options.temp = join_path(options.workspace, options.temp)

    options.merge_xunit_xmls = None if options.merge_xunit_xmls is None else [join_path(options.workspace, path) for path in options.merge_xunit_xmls]
    options.to = None if options.to is None else join_path(options.workspace, options.to)

    _options.update(options.__dict__)



import logging
import sys

from . import config


class PConsole:
    def __init__(self, out):
        self.out = out

    def write(self, msg):
        self.out.write(str(msg))

    def write_line(self, msg):
        self.out.write(str(msg) + "\n")


pconsole = PConsole(sys.stdout)
pconsole_err = PConsole(sys.stderr)


class PReporter:
    def __init__(self):
        pass

    def debug(self, msg, screenshot=False):
        self.__log(logging.DEBUG, msg, screenshot)

    def info(self, msg, screenshot=False):
        self.__log(logging.INFO, msg, screenshot)

    def warn(self, msg, screenshot=False):
        self.__log(logging.WARN, msg, screenshot)

    def error(self, msg, screenshot=False):
        self.__log(logging.ERROR, msg, screenshot)

    def critical(self, msg, screenshot=False):
        self.__log(logging.CRITICAL, msg, screenshot)

    def __log(self, level, msg, screenshot):
        from . import testexecutor, screencapturer

        try:
            running_test_fixture = testexecutor.current_executor().get_property("running_test_fixture")
        except AttributeError as e:
            pconsole.write_line("[%s] %s" % (logging.getLevelName(level), msg))
        else:
            log = {"level": logging.getLevelName(level).lower(), "message": str(msg)}
            if screenshot and not config.get_option("disable_screenshot"):
                log["screenshots"] = screencapturer.take_screenshots()

            running_test_fixture.logs.append(log)

            if config.get_option("verbose"):
                # output to pconsole
                message = "[%s] %s" % (running_test_fixture.full_name, msg)
                pconsole.write_line(message)


preporter = PReporter()



import platform
from codecs import open
from os import path

from setuptools import setup

from ptest import __version__

current_dir = path.abspath(path.dirname(__file__))
# Get the long description from the relevant file
with open(path.join(current_dir, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()
with open(path.join(current_dir, "CHANGELOG"), encoding="utf-8") as f:
    long_description += "\n" + f.read()

classifiers = ["License :: OSI Approved :: Apache Software License",
               "Topic :: Software Development :: Testing",
               "Operating System :: Microsoft :: Windows",
               "Operating System :: MacOS :: MacOS X"] + [
                  ("Programming Language :: Python :: %s" % x) for x in
                  "2.7 3.4 3.5".split()]


def make_cmdline_entry_points():
    target = "ptest.main:main"
    entry_points = []
    version = platform.python_version()
    if version[0] == "2":
        entry_points.append("ptest=%s" % target)
    elif version[0] == "3":
        entry_points.append("ptest3=%s" % target)
    entry_points.append("ptest-%s=%s" % (version[:3], target))
    return entry_points


def main():
    setup(
        name="ptest",
        description="light test runner for Python",
        long_description=long_description,
        version=__version__,
        keywords="test testing framework automation python runner",
        author="Karl Gong",
        author_email="karl.gong@outlook.com",
        url="https://github.com/KarlGong/ptest",
        license="Apache",
        entry_points={
            "console_scripts": make_cmdline_entry_points(),
        },
        classifiers=classifiers,
        packages=["ptest"],
        package_data={"ptest": ["htmltemplate/*.*"]},
        zip_safe=False,
    )


if __name__ == "__main__":
    main()



class TestClassRunMode:
    SingleLine = "singleline"
    Parallel = "parallel"


class PDecoratorType:
    BeforeSuite = "BeforeSuite"
    AfterSuite = "AfterSuite"
    TestClass = "TestClass"
    BeforeClass = "BeforeClass"
    AfterClass = "AfterClass"
    BeforeGroup = "BeforeGroup"
    AfterGroup = "AfterGroup"
    Test = "Test"
    BeforeMethod = "BeforeMethod"
    AfterMethod = "AfterMethod"

class TestFixtureStatus:
    NOT_RUN = "not_run"
    RUNNING = "running"
    PASSED = "passed"
    SKIPPED = "skipped"
    FAILED = "failed"

TestCaseStatus = TestFixtureStatus


class TestCaseCountItem:
    PASSED = TestFixtureStatus.PASSED
    FAILED = TestFixtureStatus.FAILED
    SKIPPED = TestFixtureStatus.SKIPPED
    NOT_RUN = TestFixtureStatus.NOT_RUN
    TOTAL = "total"



import types
from functools import cmp_to_key

from .enumeration import PDecoratorType, TestFixtureStatus, TestCaseCountItem, TestClassRunMode

SECOND_MICROSECOND_CONVERSION_FACTOR = 1000000.0


class TestContainer:
    def __init__(self):
        self.start_time = None
        self.end_time = None
        self.test_cases = []

    @property
    def elapsed_time(self):
        time_delta = self.end_time - self.start_time
        return time_delta.seconds + time_delta.microseconds / SECOND_MICROSECOND_CONVERSION_FACTOR

    @property
    def status_count(self):
        status_count_dict = {
            TestCaseCountItem.TOTAL: 0,
            TestCaseCountItem.PASSED: 0,
            TestCaseCountItem.FAILED: 0,
            TestCaseCountItem.SKIPPED: 0,
            TestCaseCountItem.NOT_RUN: 0
        }
        for test_case in self.test_cases:
            status_count_dict[TestCaseCountItem.TOTAL] += 1
            status_count_dict[test_case.status] += 1
        return status_count_dict

    @property
    def pass_rate(self):
        status_count = self.status_count
        return float(status_count[TestCaseCountItem.PASSED]) * 100 / status_count[TestCaseCountItem.TOTAL]


class TestSuite(TestContainer):
    def __init__(self, name):
        TestContainer.__init__(self)
        self.test_classes = []
        self.test_class_run_groups = []
        self.name = name
        self.full_name = name
        self.before_suite = BeforeSuite(self, None)
        self.after_suite = AfterSuite(self, None)

    def init(self):
        self.init_test_fixtures()
        self.init_test_class_run_groups()
        self.sort_test_class_run_groups()

    def init_test_fixtures(self):
        # reflect the before suite and after suite
        for test_class in self.test_classes:
            test_class_ref = test_class.test_class_ref.__class__()
            for element in dir(test_class_ref):
                attr = getattr(test_class_ref, element)
                if hasattr(attr, "__enabled__") and attr.__enabled__ \
                        and hasattr(attr, "__pd_type__"):
                    if attr.__pd_type__ == PDecoratorType.BeforeSuite:
                        self.before_suite = BeforeSuite(self, attr)
                    elif attr.__pd_type__ == PDecoratorType.AfterSuite:
                        self.after_suite = AfterSuite(self, attr)

    def init_test_class_run_groups(self):
        run_groups = {}
        run_group_index = 0
        for test_class in self.test_classes:
            if test_class.run_group is None:
                run_groups[run_group_index] = [test_class]
                run_group_index += 1
            elif test_class.run_group in run_groups:
                run_groups[test_class.run_group].append(test_class)
            else:
                run_groups[test_class.run_group] = [test_class]
        self.test_class_run_groups = run_groups.values()

    def sort_test_class_run_groups(self):
        run_groups = []
        # sort the test classes in run group by its run mode
        for run_group in self.test_class_run_groups:
            run_groups.append(sorted(run_group, key=lambda test_class: test_class.run_mode, reverse=True))

        # sort the test class run groups by its size and number of singleline test classes
        def cmp_run_group(run_group_a, run_group_b):
            if not len(run_group_a) == len(run_group_b):
                return len(run_group_a) - len(run_group_b)
            else:
                return len([test_class for test_class in run_group_a if test_class.run_mode == TestClassRunMode.SingleLine]) - \
                       len([test_class for test_class in run_group_b if test_class.run_mode == TestClassRunMode.SingleLine])

        self.test_class_run_groups = sorted(run_groups, key=cmp_to_key(cmp_run_group), reverse=True)

    def get_failed_setup_fixture(self):
        if self.before_suite.status == TestFixtureStatus.FAILED:
            return self.before_suite
        return None

    def get_test_class(self, full_name):
        for test_class in self.test_classes:
            if test_class.full_name == full_name:
                return test_class
        return None

    def add_test_case(self, test_class_cls, test_case_func):
        # for the @TestClass can be inherited, so set full name here
        test_class_cls.__full_name__ = "%s.%s" % (test_class_cls.__module__, test_class_cls.__name__)
        test_class = self.get_test_class(test_class_cls.__full_name__)
        if test_class is None:
            test_class = TestClass(self, test_class_cls())
            self.test_classes.append(test_class)

        test_group = test_class.get_test_group(test_case_func.__group__)
        if test_group is None:
            test_group = TestGroup(test_class, test_case_func.__group__, test_class_cls())
            test_class.test_groups.append(test_group)

        test_case = test_group.get_test_case(test_case_func.__name__)
        if test_case is None:
            if hasattr(test_class_cls, test_case_func.__name__): # normal
                test_case = TestCase(test_group, getattr(test_class_cls(), test_case_func.__name__))
            else: # mocked
                test_class_ref = test_class_cls()
                mock_method = types.MethodType(test_case_func, test_class_ref)
                setattr(test_class_ref, test_case_func.__name__, mock_method)
                test_case = TestCase(test_group, mock_method)
            test_group.test_cases.append(test_case)
            test_class.test_cases.append(test_case)
            self.test_cases.append(test_case)
            return True
        return False


class TestClass(TestContainer):
    def __init__(self, test_suite, test_class_ref):
        TestContainer.__init__(self)
        self.test_suite = test_suite
        self.test_class_ref = test_class_ref
        self.test_groups = []
        self.name = test_class_ref.__class__.__name__
        self.full_name = test_class_ref.__full_name__
        self.run_mode = test_class_ref.__run_mode__
        self.run_group = test_class_ref.__run_group__
        self.description = test_class_ref.__description__
        self.custom_args = test_class_ref.__custom_args__

        self.before_class = BeforeClass(self, None)
        self.after_class = AfterClass(self, None)
        # reflect the before class and after class
        for element in dir(test_class_ref):
            attr = getattr(test_class_ref, element)
            if hasattr(attr, "__enabled__") and attr.__enabled__ \
                    and hasattr(attr, "__pd_type__"):
                if attr.__pd_type__ == PDecoratorType.BeforeClass:
                    self.before_class = BeforeClass(self, attr)
                elif attr.__pd_type__ == PDecoratorType.AfterClass:
                    self.after_class = AfterClass(self, attr)

    def get_failed_setup_fixture(self):
        setup_fixture = self.test_suite.get_failed_setup_fixture()
        if setup_fixture:
            return setup_fixture
        if self.before_class.status == TestFixtureStatus.FAILED:
            return self.before_class
        return None

    def get_test_group(self, name):
        for test_group in self.test_groups:
            if test_group.name == name:
                return test_group
        return None

    @property
    def is_group_feature_used(self):
        return not (len(self.test_groups) == 1 and self.test_groups[0].name == "DEFAULT" and self.test_groups[
            0].before_group.is_empty and self.test_groups[0].after_group.is_empty)


class TestGroup(TestContainer):
    def __init__(self, test_class, name, test_class_ref):
        TestContainer.__init__(self)
        self.test_class = test_class
        self.test_suite = self.test_class.test_suite
        self.test_class_ref = test_class_ref
        self.test_cases = []
        self.name = name
        self.full_name = "%s<%s>" % (test_class.full_name, name)

        self.before_group = BeforeGroup(self, None)
        self.after_group = AfterGroup(self, None)
        # reflect the before group and after group
        for element in dir(test_class_ref):
            attr = getattr(test_class_ref, element)
            if hasattr(attr, "__enabled__") and attr.__enabled__ \
                    and hasattr(attr, "__group__") and attr.__group__ == self.name \
                    and hasattr(attr, "__pd_type__"):
                if attr.__pd_type__ == PDecoratorType.BeforeGroup:
                    self.before_group = BeforeGroup(self, attr)
                elif attr.__pd_type__ == PDecoratorType.AfterGroup:
                    self.after_group = AfterGroup(self, attr)

    def get_failed_setup_fixture(self):
        setup_fixture = self.test_class.get_failed_setup_fixture()
        if setup_fixture:
            return setup_fixture
        if self.before_group.status == TestFixtureStatus.FAILED:
            return self.before_group
        return None

    def get_test_case(self, name):
        for test_case in self.test_cases:
            if test_case.name == name:
                return test_case
        return None


class TestCase:
    def __init__(self, test_group, test_case_ref):
        self.test_group = test_group
        self.test_class = self.test_group.test_class
        self.test_suite = self.test_class.test_suite
        self.test_case_ref = test_case_ref
        self.name = test_case_ref.__name__
        self.full_name = "%s.%s" % (self.test_class.full_name, self.name)
        self.start_time = None
        self.end_time = None

        self.test = Test(self, test_case_ref)

        self.tags = self.test.tags
        self.expected_exceptions = self.test.expected_exceptions
        self.parameters = self.test.parameters
        self.data_index = self.test.data_index
        self.group = self.test.group
        self.description = self.test.description
        self.custom_args = self.test.custom_args
        self.location = self.test.location

        self.before_method = BeforeMethod(self, None)
        self.after_method = AfterMethod(self, None)
        # reflect the before method and after method
        for element in dir(test_case_ref.__self__):
            attr = getattr(test_case_ref.__self__, element)
            if hasattr(attr, "__enabled__") and attr.__enabled__ \
                    and hasattr(attr, "__group__") and attr.__group__ == self.group \
                    and hasattr(attr, "__pd_type__"):
                if attr.__pd_type__ == PDecoratorType.BeforeMethod:
                    self.before_method = BeforeMethod(self, attr)
                elif attr.__pd_type__ == PDecoratorType.AfterMethod:
                    self.after_method = AfterMethod(self, attr)

    def get_failed_setup_fixture(self):
        setup_fixture = self.test_group.get_failed_setup_fixture()
        if setup_fixture:
            return setup_fixture
        if self.before_method.status == TestFixtureStatus.FAILED:
            return self.before_method
        return None

    @property
    def failure_message(self):
        return self.test.failure_message

    @property
    def failure_type(self):
        return self.test.failure_type

    @property
    def stack_trace(self):
        return self.test.stack_trace

    @property
    def skip_message(self):
        return self.test.skip_message

    @property
    def status(self):
        return self.test.status

    @property
    def elapsed_time(self):
        time_delta = self.end_time - self.start_time
        return time_delta.seconds + time_delta.microseconds / SECOND_MICROSECOND_CONVERSION_FACTOR


class TestFixture:
    def __init__(self, context, test_fixture_ref, fixture_type):
        self.context = context
        self.fixture_type = fixture_type
        self.status = TestFixtureStatus.NOT_RUN
        if test_fixture_ref is None:
            self.is_empty = True
            return
        self.is_empty = False
        self.test_fixture_ref = test_fixture_ref
        self.name = test_fixture_ref.__name__
        self.failure_message = ""
        self.failure_type = ""
        self.stack_trace = ""
        self.skip_message = ""
        self.start_time = None
        self.end_time = None
        self.logs = []
        self.description = test_fixture_ref.__description__
        self.timeout = test_fixture_ref.__timeout__
        self.custom_args = test_fixture_ref.__custom_args__
        self.location = test_fixture_ref.__location__
        self.arguments_count = test_fixture_ref.__arguments_count__

    @property
    def elapsed_time(self):
        time_delta = self.end_time - self.start_time
        seconds = time_delta.seconds + time_delta.microseconds / SECOND_MICROSECOND_CONVERSION_FACTOR
        return seconds


class BeforeSuite(TestFixture):
    def __init__(self, test_suite, test_fixture_ref):
        TestFixture.__init__(self, test_suite, test_fixture_ref, PDecoratorType.BeforeSuite)
        self.test_suite = self.context
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_suite.name, self.fixture_type)


class BeforeClass(TestFixture):
    def __init__(self, test_class, test_fixture_ref):
        TestFixture.__init__(self, test_class, test_fixture_ref, PDecoratorType.BeforeClass)
        self.test_class = self.context
        self.test_suite = self.test_class.test_suite
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_class.full_name, self.fixture_type)


class BeforeGroup(TestFixture):
    def __init__(self, test_group, test_fixture_ref):
        TestFixture.__init__(self, test_group, test_fixture_ref, PDecoratorType.BeforeGroup)
        self.test_group = self.context
        self.test_class = self.test_group.test_class
        self.test_suite = self.test_group.test_suite
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_group.full_name, self.fixture_type)
            self.group = test_fixture_ref.__group__


class BeforeMethod(TestFixture):
    def __init__(self, test_case, test_fixture_ref):
        TestFixture.__init__(self, test_case, test_fixture_ref, PDecoratorType.BeforeMethod)
        self.test_case = self.context
        self.test_group = self.test_case.test_group
        self.test_class = self.test_case.test_class
        self.test_suite = self.test_case.test_suite
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_case.full_name, self.fixture_type)
            self.group = test_fixture_ref.__group__


class Test(TestFixture):
    def __init__(self, test_case, test_fixture_ref):
        TestFixture.__init__(self, test_case, test_fixture_ref, PDecoratorType.Test)
        self.full_name = "%s@%s" % (test_case.full_name, self.fixture_type)
        self.test_case = self.context
        self.test_group = self.test_case.test_group
        self.test_class = self.test_case.test_class
        self.test_suite = self.test_case.test_suite
        self.tags = test_fixture_ref.__tags__
        self.expected_exceptions = test_fixture_ref.__expected_exceptions__
        self.parameters = test_fixture_ref.__parameters__
        self.data_index = test_fixture_ref.__data_index__
        self.group = test_fixture_ref.__group__


class AfterMethod(TestFixture):
    def __init__(self, test_case, test_fixture_ref):
        TestFixture.__init__(self, test_case, test_fixture_ref, PDecoratorType.AfterMethod)
        self.test_case = self.context
        self.test_group = self.test_case.test_group
        self.test_class = self.test_case.test_class
        self.test_suite = self.test_case.test_suite
        self.always_run = False
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_case.full_name, self.fixture_type)
            self.always_run = test_fixture_ref.__always_run__
            self.group = test_fixture_ref.__group__


class AfterGroup(TestFixture):
    def __init__(self, test_group, test_fixture_ref):
        TestFixture.__init__(self, test_group, test_fixture_ref, PDecoratorType.AfterGroup)
        self.test_group = self.context
        self.test_class = self.test_group.test_class
        self.test_suite = self.test_group.test_suite
        self.always_run = False
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_group.full_name, self.fixture_type)
            self.always_run = test_fixture_ref.__always_run__
            self.group = test_fixture_ref.__group__


class AfterClass(TestFixture):
    def __init__(self, test_class, test_fixture_ref):
        TestFixture.__init__(self, test_class, test_fixture_ref, PDecoratorType.AfterClass)
        self.test_class = self.context
        self.test_suite = self.test_class.test_suite
        self.always_run = False
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_class.full_name, self.fixture_type)
            self.always_run = test_fixture_ref.__always_run__


class AfterSuite(TestFixture):
    def __init__(self, test_suite, test_fixture_ref):
        TestFixture.__init__(self, test_suite, test_fixture_ref, PDecoratorType.AfterSuite)
        self.test_suite = self.context
        self.always_run = False
        if not self.is_empty:
            self.full_name = "%s@%s" % (test_suite.name, self.fixture_type)
            self.always_run = test_fixture_ref.__always_run__


default_test_suite = TestSuite("DefaultSuite")



from ptest.decorator import TestClass, Test, BeforeMethod, AfterMethod
from ptest.assertion import assert_equals
from ptest.plogger import preporter

CN_GROUP = "CN"
US_GROUP = "US"


@TestClass()
class PTestClass:
    @BeforeMethod(group=CN_GROUP)
    def before_cn(self):
        self.expected = "cn"

    @BeforeMethod(group=US_GROUP)
    def before_us(self):
        self.expected = "us"

    @Test(group=CN_GROUP)
    def test_cn(self):
        assert_equals("cn", self.expected)

    @Test(group=US_GROUP)
    def test_us(self):
        assert_equals("us", self.expected)

    @AfterMethod(group=CN_GROUP)
    def after_cn(self):
        preporter.info("cleaning up")

    @AfterMethod(group=US_GROUP)
    def after_us(self):
        preporter.info("cleaning up")
